# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Helper functions for parsers
"""
from collections import deque

# This file contains helper methods and infrastructure for parsers. The class
# SectionLineParser in particular is a powerful tool that is intended to make
# parsing section based files easier. There's a fair chance it turned out to be
# a multi-tentacled Lovecraftion horror that's in charge of slightly magical
# switchboard. If it ever breaks I'm very sorry for you. Pray to your deity of
# choice and prepare your sacrificial chicken.


class SectionParser(type):
    """
    Metaclass (!) that populates the `METH_DICT` attribute of new classes. The
    contents of `METH_DICT` are set by reading the `_section_names` attribute
    of all its attributes. You can conveniently set `_section_names` attributes
    using the :meth:`section_parser` decorator.
    """
    def __new__(mcs, name, bases, attrs, **kwargs):
        obj = super().__new__(mcs, name, bases, attrs, **kwargs)
        #if not hasattr(obj, 'METH_DICT'):
        obj.METH_DICT = {}
        mapping = obj.METH_DICT

        for attribute_name in dir(obj):
            attribute = getattr(obj, attribute_name)
            try:
                section_names = attribute._section_names
            except AttributeError:
                pass
            else:
                for names, kwargs in section_names.items():
                    mapping[names] = (attribute, kwargs)
        return obj

    @staticmethod
    def section_parser(*names, **kwargs):
        """
        Parameters
        ----------
        names: tuple[collections.abc.Hashable]
            The section names that should be associated with the decorated
            function.
        kwargs: dict[str]
            The keyword arguments with which the decorated function should be
            called.
        """
        def wrapper(method):
            if not hasattr(method, '_section_names'):
                method._section_names = {}
            method._section_names[names] = kwargs
            return method
        return wrapper


class LineParser:
    """
    Class that describes a parser object that parses a file line by line.
    Subclasses will probably want to override the methods :meth:`dispatch`,
    :meth:`parse_line`, and/or :meth:`finalize`:

      - :meth:`dispatch` is called for every line and should return the
        function that should be used to parse that line.
      - :meth:`parse_line` is called by the default implementation of
        :meth:`dispatch` for every line.
      - :meth:`finalize` is called at the end of the file.
    """
    COMMENT_CHAR = '#'

    def parse(self, file_handle):
        """
        Reads lines from `file_handle`, and calls :meth:`dispatch` to find
        which method to call to do the actual parsing. Yields the result of
        that call, if it's not `None`.
        At the end, calls :meth:`finalize`, and yields its results, iff
        it's not None.

        Parameters
        ----------
        file_handle: collections.abc.Iterable[str]
            The data to parse. Should produce lines of data.

        Yields
        ------
        object
            The results of dispatching to parsing methods, and of
            :meth:`finalize`.
        """
        lineno = 0
        for lineno, line in enumerate(file_handle, 1):
            line, _ = split_comments(line, self.COMMENT_CHAR)
            if not line:
                continue
            result = self.dispatch(line)(line, lineno)

            if result is not None:
                yield result

        result = self.finalize(lineno)
        if result is not None:
            yield result

    def finalize(self, lineno=0):
        """
        Wraps up. Is called at the end of the file.
        """
        return

    def dispatch(self, line):
        """
        Finds the correct method to parse `line`. Always returns
        :meth:`parse_line`.
        """
        return self.parse_line

    def parse_line(self, line, lineno):
        """
        Does nothing and should be overridden by subclasses.
        """
        return


class SectionLineParser(LineParser, metaclass=SectionParser):
    """
    Baseclass for all parsers that have to parse file formats that are based on
    sections. Parses the `macros` section.
    Subclasses will probably want to override :meth:`finalize` and/or
    :meth:`finalize_section`.

    :meth:`finalize_section` is called with the previous section whenever a
    section ends.

    Attributes
    ----------
    section: list[str]
        The current section.
    macros: dict[str, str]
        A set of subsitution rules as parsed from a `macros` section.
    """
    METH_DICT = {}
    """
    A dict of all known parser methods, mapping section names to the
    function to be called and the associated keyword arguments.
    """

    def __init__(self, *args, **kwargs):
        self.macros = {}
        self.section = []
        super().__init__(*args, **kwargs)

    def dispatch(self, line):
        """
        Looks at `line` to see what kind of line it is, and returns either
        :meth:`parse_header` if `line` is a section header or
        :meth:`parse_section` otherwise. Calls :meth:`is_section_header` to see
        whether `line` is a section header or not.

        Parameters
        ----------
        line: str

        Returns
        -------
        collections.abc.Callable
            The method that should be used to parse `line`.
        """
        if self.is_section_header(line):
            return self.parse_header
        else:
            return self.parse_section

    def finalize(self, lineno=0):
        """
        Called after the last line has been parsed to wrap up. Resets
        the instance and calls :meth:`finalize_section`.

        Arguments
        ---------
        lineno: int
            The line number.
        """
        prev_section = self.section
        self.section = []
        result = self.finalize_section(prev_section, prev_section)
        self.macros = {}
        self.section = None
        return result

    def finalize_section(self, previous_section, ended_section):
        """
        Called once a section is finished. Currently does nothing.

        Arguments
        ---------
        previous_section: list[str]
            The last parsed section.
        ended_section: list[str]
            The sections that have been ended.
        """
        return

    def parse_section(self, line, lineno):
        """
        Parse `line` with line number `lineno` by looking up the section in
        :attr:`METH_DICT` and calling that method.

        Parameters
        ----------
        line: str
        lineno: int

        Returns
        -------
        object
            The result returned by calling the registered method.
        """
        line = _substitute_macros(line, self.macros)
        if tuple(self.section) not in self.METH_DICT:
            raise IOError("Can't parse line {} in section '{}' because the "
                          "section is unknown".format(lineno, self.section))
        try:
            method, kwargs = self.METH_DICT[tuple(self.section)]
            return method(self, line, lineno, **kwargs)
        except Exception as error:
            raise IOError("Problems parsing line {}. I think it should be a "
                          "'{}' line, but I can't parse it as such."
                          "".format(lineno, self.section)) from error

    def parse_header(self, line, lineno=0):
        """
        Parses a section header with line number `lineno`. Sets :attr:`section`
        when applicable. Does not check whether `line` is a valid section
        header.

        Parameters
        ----------
        line: str
        lineno: str

        Returns
        -------
        object
            The result of calling :meth:`finalize_section`, which is called
            if a section ends.

        Raises
        ------
        KeyError
            If the section header is unknown.
        """
        prev_section = self.section

        section = self.section + [line.strip('[ ]').casefold()]

        ended = []
        while tuple(section) not in self.METH_DICT and len(section) > 1:
            ended.append(section.pop(-2))  # [a, b, c, d] -> [a, b, d]

        self.section = section
        if prev_section:
            result = self.finalize_section(prev_section, ended)
            return result

    @staticmethod
    def is_section_header(line):
        """
        Parameters
        ----------
        line: str
            A line of text.

        Returns
        -------
        bool
            ``True`` iff `line` is a section header.

        Raises
        ------
        IOError
            The line starts like a section header but looks misformatted.
        """
        if line.startswith('['):
            if line.endswith(']'):
                return True
            else:
                raise IOError('Section header looks misformatted.')
        return False

    @SectionParser.section_parser('macros')
    def _macros(self, line, lineno=0):
        """
        Parses a "macros" section. Adds to :attr:`macros`.

        Parameters
        ----------
        line: str
        """
        line = deque(_tokenize(line))
        _parse_macro(line, self.macros)


def split_comments(line, comment_char=';'):
    """
    Splits `line` at the first occurence of `comment_char`.

    Parameters
    ----------
    line: str
    comment_char: str

    Returns
    -------
    tuple[str, str]
        `line` before and after `comment_char`, respectively. If `line` does
        not contain `comment_char`, the second element will be an empty string.
    """
    split = line.split(comment_char, 1)
    data = split[0].strip()
    if len(split) == 1:
        return data, ''
    else:
        return data, split[1].strip()


def _tokenize(line):
    """
    Split an interaction line into its elementary components.

    An interaction line is any uncommented and non empty line that follows a
    section header about an interaction type. Such a line is composed of the
    following parts:

    * a list of atoms involved in the interaction,
    * an optional delimiter that indicates the end of the atom list,
    * a list of parameters for the interaction.

    The list of atoms is *a minima* a list of atom references. In blocks, these
    references can be atom 1-based indices referring to the order of the atoms
    in the "[ atoms ]" section. It is however more readable, and more robust,
    to refer to atoms by their name. Only the reference by name is allowed in
    links, as links may not have a full "[ atoms ]" section. In links, each
    atom reference can be complemented by atom attributes to specify the  scope
    of the link. These attribute follow the atom reference and are formatted
    like a python dictionary.

    The end-of-atoms delimiter is useful for interaction types that are not
    explicitly encoded in the parser. It allows to indicate when the list of
    atoms ends, and where the list of parameters starts. Two dashes ("--") are
    used as the delimiter. The delimiter is optional for the interaction types
    that are explicitly encoded in the parser and that refer to a fixed number
    of atoms.

    The list of parameters will be copied as-is in an ITP file.

    In its simplest form, an interaction line is what is used in an ITP file.
    Here is an example for a bond:

        2  3  1  0.2  1000

    The two first numbers refer to the second and third atoms of the block,
    respectively. The next three values are the parameters for a bond (*i.e.*
    the function type, the equilibrium distance, and the force constant).

    The two first numbers could be replaced by the corresponding atom names:

        PO4  GL1  1 0.2  1000

    where "PO4" and "GL1" are the names of the second and third atoms of the
    block.

    Optionally, the "--" delimiter can be used after the list of atoms:

        PO4  GL1  --  1 0.2  1000

    If the line is part of a link, then the atom selection may be limited in
    scope. Atom attributes is how to implement such scope limitation:

        BB {'resname': 'ALA', 'secstruc': 'H'} BB {'resname': 'LYS', 'secstruc': 'H', 'order': +1} 1 0.2 1000

    Here, we add a bond to the current link. At one end of the bond is the atom
    named "BB" and annotated as part of an alpha helix ('secstruc': 'H') of a
    residue called "ALA". On the other end of the link is an other
    atom named "BB" that is part of an alpha helix, but that is part of the
    next residue ('order': +1) if this next residue is named "LYS".

    The order parameter has a shortcut in the form of a + or - prefix to the
    atom reference name. Then, "+ATOM" refers to "ATOM" in the next residue,
    and is equivalent to "ATOM {'order': +1}"; "-ATOM" refers to the previous
    residue. There can be multiple + or -, "++ATOM" is equivalent to "ATOM
    {'order': +2}".

    When using attributes, the optional delimiter can increase the readability:

        BB {'resname': 'ALA', 'secstruc': 'H'} +BB {'resname': 'LYS', 'secstruc': 'H'} -- 1 0.2 1000

    Tokens on an interaction line are its different elements. These elements
    are considered as one token each: am atom reference, a set of atom
    attributes, the optional delimiter, each space-separated element of the
    parameter list. The line above splits into the following tokens:

    * ``BB``
    * ``{'resname': 'ALA', 'secstruc': 'H'}``
    * ``+BB``
    * ``{'resname': 'LYS', 'secstruc': 'H'}``
    * ``--``
    * ``1``
    * ``0.2``
    * ``1000``

    Atom attributes can be written next to the previous or the next token
    without an explicit separator. The two following lines yield the same three
    tokens:

        ATOM1{attributes}ATOM2
        ATOM1 {attributes} ATOM2

    Parameters
    ----------
    line: str

    Returns
    -------
    list of str
    """
    separators = ' \t\n'
    tokens = []
    start = 0
    end = -1
    # Find the first non-separator character
    for start, char in enumerate(line):
        if char not in separators:
            break

    # Find the tokens. This has to be a while-loop because we cannot predict
    # what will be the next value of start.
    while start < len(line):
        end = start
        # We count the brackets because if a token starts with an opening
        # bracket, we want to end it with the *matching* closing bracket.
        # Note also that we do not yet implement a way to escape a bracket, nor
        # do we check if the bracket is not part of a string.
        brackets = 0
        for end, end_char in enumerate(line[start:], start=start):
            if end_char == '{':
                # We reached an opening bracket. If it is the first character
                # of the token or if we are already engaged in a bracketized
                # token, then we go on. But if the current token was not
                # a bracketized token, it means we are at the beginning of
                # a new token, so we treat the opening bracket as a separator.
                if not brackets and end != start:
                    end -= 1
                    break
                brackets += 1
            elif end_char == '}':
                brackets -= 1
                if not brackets:
                    break
            elif end_char in separators:
                if not brackets:
                    # We reached a separator. We do not want the separator to
                    # be included in the token, so we push the end by one
                    # character to the left.
                    end -= 1
                    break
        if brackets > 0:
            msg = 'Unexpected end of line. A closing bracket is missing.'
            raise IOError(msg)
        elif brackets < 0:
            msg = 'An opening bracket is missing.'
            raise IOError(msg)

        token = line[start:end + 1]
        if token:
            tokens.append(token)

        # Find the beginning of the next token.
        start = end + 1
        while start < len(line) and line[start] in separators:
            start += 1
    return tokens


def _substitute_macros(line, macros):
    r"""
    Substitute macros by their content.

    A macro starts with a '$' and ends with one amongst ' ${}\n\t"'.

    Parameters
    ----------
    line: str
        The line to fix.
    macros: dict[str, str]
        Keys are macro names, values are the replacement content.

    Returns
    -------
    str
    """
    start = None
    while True:  # stops when start < 0
        start = line.find('$', start)
        if start < 0:
            break
        for end, char in enumerate(line[start + 1:], start=start + 1):
            if char in ' \t\n{}$"':
                break
        else: # no break
            end += 1
        macro_name = line[start + 1:end]
        macro_value = macros[macro_name]
        line = line[:start] + macro_value + line[end:]
        end = start + len(macro_value)
    return line


def _parse_macro(tokens, macros):
    if len(tokens) > 2:
        raise IOError('Unexpected column in macro definition.')
    elif len(tokens) < 2:
        raise IOError('Missing column in macro definition.')
    macro_name = tokens.popleft()
    macro_value = tokens.popleft()
    macros[macro_name] = macro_value



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Assign protein secondary structures using DSSP.
"""

import collections
import logging
import os
import subprocess
import tempfile

from ..file_writer import open
from ..pdb import pdb
from ..system import System
from ..processors.processor import Processor
from ..selectors import is_protein, selector_has_position, filter_minimal, select_all
from .. import utils
from ..log_helpers import StyleAdapter, get_logger

LOGGER = StyleAdapter(get_logger(__name__))


class DSSPError(Exception):
    """
    Exception raised if DSSP fails.
    """


def read_dssp2(lines):
    """
    Read the secondary structure from a DSSP output.

    Only the first column of the "STRUCTURE" block is read. See the
    `documentation of the DSSP format`_ for more details.

    The secondary structures that can be read are:

    :H: α-helix
    :B: residue in isolated β-bridge
    :E: extended strand, participates in β ladder
    :G: 3-helix (3-10 helix)
    :I: 5 helix (π-helix)
    :T: hydrogen bonded turn
    :S: bend
    :C: loop or irregular

    The "C" code for loops and random coil is translated from the gap used in
    the DSSP file for an improved readability.

    Only the version 2 and 3 of DSSP is supported. If the format is not
    recognized as comming from that version of DSSP, then a :exc:`IOError` is
    raised.

    .. _`documentation of the DSSP format`: http://swift.cmbi.ru.nl/gv/dssp/DSSP_3.html

    Parameters
    ----------
    lines:
        An iterable over the lines of the DSSP output. This can be *e.g.* a
        list of lines, or a file handler. The new line character is ignored.

    Returns
    -------
    secstructs: list[str]
        The secondary structure assigned by DSSP as a list of one-letter
        secondary structure code.

    Raises
    ------
    IOError
        When a line could not be parsed, or if the version of DSSP
        is not supported.
    """
    secstructs = []
    # We use the line number for the error messages. It is more natural for a
    # user to count lines in a file starting from 1 rather than 0.
    numbered_lines = enumerate(lines, start=1)

    # The function can only read output from DSSP version 2 and 3. Hopefully, if
    # the input file is not in this format, then the parser will break as it
    # reads the file; we can expect that the end of the header will not be found
    # or the secondary structure will be an unexpected character.
    # We could predict from the first line that the format is not the one we
    # expect if it does not start with "===="; however, the first lines of the
    # file are non-essential and could have been trimmed. (For instance, the
    # first line of a DSSPv2 file contains the date of execution of the
    # program, which is annoying when comparing files.) Failing at the
    # first line is likely unnecessary.
    # Yet, we can identify files from DSSP v1 from the first line. These files
    # start with "****" instead of "====". If we identify such a file, we can
    # fail with a useful error message.
    _, first_line = next(numbered_lines)
    if first_line and first_line.startswith('****'):
        msg = ('Based on its header, the input file could come from a '
               'pre-July 1995 version of DSSP (or the compatibility mode '
               'of a more recent version). Only output from the version 2 and 3'
               'of DSSP are supported.')
        raise IOError(msg)

    # First we skip the header and the histogram.
    for line_num, line in numbered_lines:
        if line.startswith('  #  RESIDUE AA'):
            break
    else:  # no break
        msg = ('No secondary structure assignation could be read because the '
               'file is not formated correctly. No line was found that starts '
               'with "  #  RESIDUE AA".')
        raise IOError(msg)

    # Now, every line should be a secondary structure assignation.
    for line_num, line in numbered_lines:
        if '!' in line or not line:
            # This is a TER record or an empty line, we ignore it.
            continue
        elif len(line) >= 17:
            secondary_structure = line[16]
            if secondary_structure not in 'HBEGITS ':
                msg = 'Unrecognize secondary structure "{}" in line {}: "{}"'
                raise IOError(msg.format(secondary_structure, line_num, line))
            # DSSP represents the coil with a space. While this works in a
            # column based file, it is much less convenient to handle in
            # our code, and it is much less readable in our debug logs.
            # We translate the space to "C" in our representation.
            if secondary_structure == ' ':
                secondary_structure = 'C'
            secstructs.append(secondary_structure)
        else:
            raise IOError('Line {} is too short: "{}".'.format(line_num, line))

    return secstructs


def run_dssp(system, executable='dssp', savefile=None):
    """
    Run DSSP on a system and return the assigned secondary structures.

    Run DSSP using the path (or name in the research PATH) given by
    "executable". Return the secondary structure parsed from the output of the
    program.

    In order to call DSSP, a PDB file is produced. Therefore, all the molecules
    in the system must contain the required attributes for such a file to be
    generated. Also, the atom names are assumed to be compatible with the
    'universal' force field for DSSP to recognize them.
    However, the molecules do not require the edges to be defined.

    DSSP is assumed to be in version 2 or 3. The secondary structure codes are
    described in :func:`read_dssp2`.

    If "savefile" is set to a path, then the output of DSSP is written in
    that file.

    Parameters
    ----------
    system: System
    executable: str
        Where to find the DSSP executable.
    savefile: None or str or pathlib.Path
        If set to a path, the output of DSSP is written in that file.

    Returns
    list of str
        The assigned secondary structures as a list of one-letter codes.
        The secondary structure sequences of all the molecules are combined
        in a single list without delimitation.

    Raises
    ------
    DSSPError
        DSSP failed to run.
    IOError
        The output of DSSP could not be parsed.

    See Also
    --------
    read_dssp2
        Parse a DSSP output.
    """
    tmpfile_handle, tmpfile_name = tempfile.mkstemp(suffix='.pdb', text=True,
                                                    dir='.', prefix='dssp_in_')
    tmpfile_handle = os.fdopen(tmpfile_handle, mode='w')
    tmpfile_handle.write(pdb.write_pdb_string(system, conect=False))
    tmpfile_handle.close()

    process = subprocess.run(
        [executable, '-i', tmpfile_name],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=False,
        universal_newlines=True
    )

    status = process.returncode
    # If an error is encountered, or the loglevel is low enough, preserve the
    # DSSP input file, and print a nice message.
    if not status and LOGGER.getEffectiveLevel() > logging.DEBUG:
        os.remove(tmpfile_name)
    if status:
        message = 'DSSP encountered an error. The message was {err}. The input' \
                  ' file provided to DSSP can be found at {file}.'
        raise DSSPError(message.format(err=process.stderr, file=tmpfile_name))
    else:
        LOGGER.debug('DSSP input file written to {}', tmpfile_name)

    if savefile is not None:
        with open(str(savefile), 'w') as outfile:
            outfile.write(process.stdout)
    return read_dssp2(process.stdout.split('\n'))


def _savefile_path(molecule, savedir=None):
    savefile = None
    if savedir is not None:
        first_atom = list(molecule.nodes.keys())[0]
        chain = molecule.nodes[first_atom].get('chain')
        if chain is None:
            msg = 'The "savedir" argument can only be used if chains are set.'
            raise ValueError(msg)
        savefile = os.path.join(savedir, 'chain_{}.ssd'.format(chain))
    return savefile


def annotate_dssp(molecule, executable='dssp', savedir=None, attribute='secstruct'):
    """
    Adds the DSSP assignation to the atoms of a molecule.

    Runs DSSP on the molecule and adds the secondary structure assignation as
    an attribute of its atoms. The attribute name in which the assignation is
    stored is controlled with the "attribute" argument.

    Only proteins can be annotated. Non-protein molecules are returned
    unmodified, so are empty molecules, and molecules for which no positions
    are set.

    The atom names are assumed to be compatible with DSSP. Atoms with no known
    position are not passed to DSSP which may lead to an error in DSSP.

    .. warning::

        The molecule is annotated **in-place**.

    Parameters
    ----------
    molecule: Molecule
        The molecule to annotate. Its atoms must have the attributes required
        to write a PDB file; other atom attributes, edges, or molecule
        attributes are not used.
    executable: str
        The path or name in the research PATH of the DSSP executable.
    savedir: None or str
        If set to a path, the DSSP output will be written in this **directory**.
        The option is only available if chains are defined with the 'chain'
        atom attribute.
    attribute: str
        The name of the atom attribute in which to store the annotation.

    See Also
    --------
    run_dssp, read_dssp2
    """
    if not is_protein(molecule):
        return

    clean_pos = molecule.subgraph(
        filter_minimal(molecule, selector=selector_has_position)
    )

    # We ignore empty molecule, there is no point at running DSSP on them.
    if not clean_pos:
        return

    savefile = _savefile_path(molecule, savedir)

    system = System()
    system.add_molecule(clean_pos)
    secstructs = run_dssp(system, executable, savefile)

    annotate_residues_from_sequence(molecule, attribute, secstructs)


def convert_dssp_to_martini(sequence):
    """
    Convert a sequence of secondary structure to martini secondary sequence.

    Martini treats some secondary structures with less resolution than dssp.
    For instance, the different types of helices that dssp discriminates are
    seen the same by martini. Yet, different parts of the same helix are seen
    differently in martini.

    In the Martini force field, the B and E secondary structures from DSSP are
    both treated as extended regions. All the DSSP helices are treated the
    same, but the different part of the helices (beginning, end, core of a
    short helix, core of a long helix) are treated differently.

    After the conversion, the secondary structures are:
    * :F: Collagenous Fiber
    * :E: Extended structure (β sheet)
    * :H: Helix structure
    * :1: Helix start (H-bond donor)
    * :2: Helix end (H-bond acceptor)
    * :3: Ambivalent helix type (short helices)
    * :T: Turn
    * :S: Bend
    * :C: Coil

    Parameters
    ----------
    sequence: str
        A sequence of secondary structures as read from dssp. One letter per
        residue.

    Returns
    -------
    str
        A sequence of secondary structures usable for martini. One letter per
        residue.
    """
    ss_cg = {'1': 'H', '2': 'H', '3': 'H', 'H': 'H', 'G': 'H', 'I': 'H',
             'B': 'E', 'E': 'E', 'T': 'T', 'S': 'S', 'C': 'C'}
    patterns = collections.OrderedDict([
        ('.H.', '.3.'), ('.HH.', '.33.'), ('.HHH.', '.333.'),
        ('.HHHH.', '.3333.'), ('.HHHHH.', '.13332.'),
        ('.HHHHHH.', '.113322.'), ('.HHHHHHH.', '.1113222.'),
        ('.HHHH', '.1111'), ('HHHH.', '2222.'),
    ])
    cg_sequence = ''.join(ss_cg[secstruct] for secstruct in sequence)
    wildcard_sequence = ''.join('H' if secstruct == 'H' else '.'
                                for secstruct in cg_sequence)
    for pattern, replacement in patterns.items():
        wildcard_sequence = wildcard_sequence.replace(pattern, replacement)
    result = ''.join(
        wildcard if wildcard != '.' else cg
        for wildcard, cg in zip(wildcard_sequence, cg_sequence)
    )
    return result


def sequence_from_residues(molecule, attribute, default=None):
    """
    Generates a sequence of `attribute`, one per residue in `molecule`.

    Parameters
    ----------
    molecule: vermouth.molecule.Molecule
        The molecule to process.
    attribute: collections.abc.Hashable
        The attribute of interest.
    default: object
        Yielded if the first node of a residue has no attribute `attribute`.

    Yields
    ------
    object
        The value of `attribute` for every residue in `molecule`.
    """
    for residue_nodes in molecule.iter_residues():
        # TODO: Make sure they're the same for every node in residue.
        first_name = residue_nodes[0]
        first_node = molecule.nodes[first_name]
        value = first_node.get(attribute, default)
        yield value


def annotate_residues_from_sequence(molecule, attribute, sequence):
    """
    Sets the attribute `attribute` to a value from `sequence` for every node in
    `molecule`. Nodes in the n'th residue of `molecule` are given the n'th
    value of `sequence`.

    Parameters
    ----------
    molecule: networkx.Graph
        The molecule to annotate. Is modified in-place.
    attribute: collections.abc.Hashable
        The attribute to set.
    sequence: collections.abc.Sequence
        The values assigned.

    Raises
    ------
    ValueError
        If the length of `sequence` is different from the number of residues in
        `molecule`.
    """
    residues = list(molecule.iter_residues())
    if len(sequence) == 1:
        sequence = sequence * len(residues)
    elif len(sequence) != len(residues):
        msg = ('The sequence length does not match the number of residues. '
               'The sequence has {} elements for {} residues.')
        raise ValueError(msg.format(len(sequence), len(residues)))
    for residue_nodes, value in zip(residues, sequence):
        for node_name in residue_nodes:
            molecule.nodes[node_name][attribute] = value


def convert_dssp_annotation_to_martini(
        molecule, from_attribute='secstruct', to_attribute='cgsecstruct'):
    """
    For every node in `molecule`, translate the `from_attribute` with
    :func:`convert_dssp_to_martini`, and assign it to the attribute
    `to_attribute`.

    Parameters
    ----------
    molecule: networkx.Graph
        The molecule to process. Is modified in-place.
    from_attribute: collections.abc.Hashable
        The attribute to read.
    to_attribute: collections.abc.Hashable
        The attribute to set.

    Raises
    ------
    ValueError
        If not all nodes have a `from_attribute`.
    """
    dssp_sequence = list(sequence_from_residues(molecule, from_attribute))
    if None not in dssp_sequence:
        cg_sequence = list(convert_dssp_to_martini(dssp_sequence))
        annotate_residues_from_sequence(molecule, to_attribute, cg_sequence)
    elif all(elem is None for elem in dssp_sequence):
        # There is no DSSP assignation for the molecule. This is likely due to
        # the molecule not being a protein. Anyway, we issue a debug message
        # as it *could* be due to the DSSP assignation having been skipped
        # for some reason.
        msg = 'No DSSP assignation to convert to Martini secondary structure intermediates.'
        LOGGER.debug(msg)
    else:
        # This is more of a problem. For now, we do not know what to do with
        # incomplete DSSP assignation. This may come later as a problem if
        # a molecule is attached to a protein.
        raise ValueError('Not all residues have a DSSP assignation.')


class AnnotateDSSP(Processor):
    name = 'AnnotateDSSP'

    def __init__(self, executable='dssp', savedir=None):
        super().__init__()
        self.executable = executable
        self.savedir = savedir

    def run_molecule(self, molecule):
        annotate_dssp(molecule, self.executable, self.savedir)
        return molecule


class AnnotateMartiniSecondaryStructures(Processor):
    name = 'AnnotateMartiniSecondaryStructures'

    @staticmethod
    def run_molecule(molecule):
        convert_dssp_annotation_to_martini(molecule)
        return molecule


class AnnotateResidues(Processor):
    """
    Set an attribute of the nodes from a sequence with one element per residue.

    Read a sequence with one element per residue and assign an attribute of
    each node based on that sequence, so each node has the value corresponding
    to its residue. In most cases, the length of the sequence has to match the
    total number of residues in the system. The sequence must be ordered in the
    same way as the residues in the system. If all the molecules have the same
    number of residues, and if the length of the sequence corresponds to the
    number of residue of one molecule, then the sequence is repeated to all
    molecules. If the sequence contains only one element, then it is repeated
    to all the residues ofthe system.

    Parameters
    ----------
    attribute: str
        Name of the node attribute to populate.
    sequence: collections.abc.Sequence
        Per-residue sequence.
    molecule_selector: collections.abc.Callable
        Function that takes an instance of :class:`vermouth.molecule.Molecule`
        as argument and returns `True` if the molecule should be considered,
        else `False`.
    """
    name = 'AnnotateResidues'

    def __init__(self, attribute, sequence,
                 molecule_selector=select_all):
        self.attribute = attribute
        self.sequence = sequence
        self.molecule_selector = molecule_selector

    def run_molecule(self, molecule):
        """
        Run the processor on a single molecule.

        Parameters
        ----------
        molecule: vermouth.molecule.Molecule

        Returns
        -------
        vermouth.molecule.Molecule
        """
        if self.molecule_selector(molecule):
            annotate_residues_from_sequence(molecule, self.attribute, self.sequence)
        return molecule

    def run_system(self, system):
        """
        Run the processor on a system.

        Parameters
        ----------
        system: vermouth.system.System

        Returns
        -------
        vermouth.system.System
        """
        # Test and adjust the length of the sequence. There are 3 valid scenarios:
        # * the length of the sequence matches the number of residues in the
        #   selection;
        # * all the molecules in the selection have the same number of residues
        #   and the sequence length matches the number of residue of one
        #   molecule; in this case the equence is repeated for each molecule;
        # * the sequence has a length of one; in this case the sequence is
        #   repeated for each residue.
        # The case were there is no molecule in the selection is only valid if
        # the sequence is empty. Then we are in the first valid scenario.
        molecule_lengths = [
            len(list(molecule.iter_residues()))
            for molecule in system.molecules
            if self.molecule_selector(molecule)
        ]
        if self.sequence and not molecule_lengths:
            raise ValueError('There is no molecule to which '
                             'to apply the sequence.')
        if (molecule_lengths
                and len(self.sequence) == molecule_lengths[0]
                and utils.are_all_equal(molecule_lengths)):
            sequence = list(self.sequence) * len(molecule_lengths)
        elif len(self.sequence) == 1:
            sequence = list(self.sequence) * sum(molecule_lengths)
        elif len(self.sequence) != sum(molecule_lengths):
            raise ValueError(
                'The length of the sequence ({}) does not match the '
                'number of residues in the selection ({}).'
                .format(len(self.sequence), sum(molecule_lengths))
            )
        else:
            sequence = self.sequence

        end = 0
        begin = 0
        for molecule, nres in zip(system.molecules, molecule_lengths):
            end += nres
            annotate_residues_from_sequence(
                molecule,
                self.attribute,
                sequence[begin:end]
            )
            begin += nres



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hypothesis.strategies as st
from hypothesis_networkx import strategy as hnst
import vermouth
import vermouth.molecule
from vermouth.molecule import Interaction, Molecule, Block, Link, DeleteInteraction


# pylint: disable=no-value-for-parameter

@st.composite
def attribute_dict(draw, min_size=0, max_size=None, max_depth=1):
    """
    Strategy that builds an attribute dictionary for meta or atoms.

    Parmeters
    ---------
    draw:
        Internal for hypothesis.
    min_size: int
        The minimum number of key-value pairs.
    max_size: int, optional
        The maximum number of key-value pairs. If set to `None` (default),
        there is no bound set for the maximum size of the dictionary in the
        same manner as :func:`hypothesis.strategies.dictionaries`.
    max_depth: int
        The strategy is recursive so that the values can be attribute
        dictionaries or lists containing attribute dictionaries. This argument
        sets the maximum depth of the recursion.

    Returns
    -------
    hypothesis.searchstrategy.lazy.LazyStrategy
    """
    keys = st.one_of(st.integers(), st.none(), st.text())
    bases = [st.none(), st.text(), st.integers(), st.floats()]
    if max_depth > 0:
        bases.append(attribute_dict(max_size=1, max_depth=max_depth - 1))
    lists = st.lists(st.one_of(*bases), max_size=2)
    values = st.one_of(*bases, lists)
    return draw(st.dictionaries(keys, values, min_size=min_size, max_size=max_size))


@st.composite
def parameter_effectors(draw):
    """
    Strategy that builds a :class:`~vermouth.molecule.LinkParameterEffector`.

    The strategy chooses one possible parameter effector class, and creates an
    instance with random keys and format spec.
    """
    possible_effectors = [
        vermouth.molecule.ParamDistance,
        vermouth.molecule.ParamAngle,
        vermouth.molecule.ParamDihedral,
        vermouth.molecule.ParamDihedralPhase,
    ]
    effector = draw(st.sampled_from(possible_effectors))
    n_keys = effector.n_keys_asked
    keys = [draw(st.text(min_size=1, max_size=4)) for _ in range(n_keys)]
    possible_formats = [None, '.2f', '3.0f']
    format_spec = draw(st.sampled_from(possible_formats))

    return effector(keys, format_spec=format_spec)


@st.composite
def random_interaction(draw, graph, natoms=None,
                       interaction_class=Interaction, attrs=False):
    """
    Strategy that builds an interaction or interaction-like object.

    By default, the strategy builds an instance of
    :class:`~vermouth.molecule.Interaction`. The strategy can also build a
    :class:`~vermouth.molecule.DeleteInteraction` by giving `DeleteInteraction`
    to the `interaction_class` argument and `True` to the `attrs` one.

    Parameters
    ----------
    draw:
        Internal for hypothesis.
    graph: networkx.Graph
        Graph/molecule from which nodes will be drawn.
    natoms: int, optional
        The number of atoms involved in the interaction. If set to `None`
        (default), a random number between 1 and 4 (included) is used.
    interaction_class: type
        The class to use for the interaction.
    attrs: bool
        Wether ot not to build an `atom_attrs` argument for the interaction.
        This should be `False` if the interaction is a
        :class:`~vermouth.molecule.Interaction`, and `True` if it is a
        :class:`~vermouth.molecule.DeleteInteraction`.

    Returns
    -------
    hypothesis.searchstrategy.lazy.LazyStrategy
    """
    if natoms is None:
        natoms = draw(st.integers(min_value=1, max_value=4))
    if graph:
        atoms = tuple(draw(st.sampled_from(list(graph.nodes))) for _ in range(natoms))
    else:
        atoms = []
    parameters = st.lists(elements=st.one_of(st.text(), parameter_effectors()))
    meta = draw(st.one_of(st.none(), attribute_dict()))
    if attrs:
        atom_attrs = tuple(draw(attribute_dict(max_size=3)) for _ in atoms)
        return interaction_class(
            atoms=atoms,
            atom_attrs=atom_attrs,
            parameters=parameters,
            meta=meta,
        )
    return interaction_class(atoms=atoms, parameters=parameters, meta=meta)


@st.composite
def interaction_collection(draw, graph,
                           interaction_class=Interaction, attrs=False):
    """
    Strategy that builds a collection of interaction-like instances.

    The collection is a dictionary with any string as key, and a list of
    :class:`~vermouth.molecule.Interaction` or
    :class:`~vermouth.molecule.DeleteInteraction`.

    The parameters are passed to :func:`random_interaction`.

    See Also
    --------
    random_interaction
    """
    result = {}
    ninteraction_types = draw(st.integers(min_value=0, max_value=2))
    for _ in range(ninteraction_types):
        ninteractions = draw(st.integers(min_value=0, max_value=2))
        type_name = draw(st.text())
        if type_name not in result:
            result[type_name] = []
        for _ in range(ninteractions):
            interaction = draw(random_interaction(
                graph, interaction_class=interaction_class, attrs=attrs,
            ))
            result[type_name].append(interaction)
    return result


@st.composite
def random_molecule(draw, molecule_class=Molecule, max_nodes=5, max_meta=None):
    """
    Strategy that builds a molecule.

    Parameters
    ----------
    draw:
        Internal for hypothesis.
    molecule_class: type
        The class of molecule to build, :class:`vermouth.molecule.Molecule` by
        default.

    Returns
    -------
    hypothesis.searchstrategy.lazy.LazyStrategy
    """
    graph = draw(hnst.graph_builder(
        max_nodes=max_nodes,
        node_data=attribute_dict(max_size=3),
        edge_data=attribute_dict(max_size=2),
    ))
    meta = draw(attribute_dict(max_size=max_meta))
    nrexcl = draw(st.one_of(st.none(), st.integers()))
    molecule = molecule_class(graph, meta=meta, nrexcl=nrexcl)

    molecule.interactions = draw(interaction_collection(molecule))

    return molecule


@st.composite
def random_block(draw, block_class=Block, max_nodes=5, max_meta=None):
    """
    Strategy that builds a block.

    Parameters
    ----------
    draw:
        Internal for hypothesis.
    block_class: type
        The class of block to build, :class:`vermouth.molecule.Block` by
        default.

    Returns
    -------
    hypothesis.searchstrategy.lazy.LazyStrategy
    """
    block = draw(random_molecule(molecule_class=block_class,
                                 max_nodes=max_nodes, max_meta=max_meta))
    block.name = draw(st.one_of(st.none(), st.text()))
    return block


@st.composite
def random_link(draw, max_nodes=5, max_meta=None):
    """
    Strategy that builds a link.

    Parameters
    ----------
    draw:
        Internal for hypothesis.
    link_class: type
        The class of block to build, :class:`vermouth.molecule.Link` by
        default.

    Returns
    -------
    hypothesis.searchstrategy.lazy.LazyStrategy
    """
    link = draw(random_block(block_class=Link,
                             max_nodes=max_nodes, max_meta=max_meta))
    link.removed_interactions = draw(interaction_collection(
        link, interaction_class=DeleteInteraction, attrs=True,
    ))
    link.molecule_meta = draw(attribute_dict(max_size=4))
    link.non_edges = draw(st.lists(
        st.tuples(st.text(min_size=1, max_size=4), attribute_dict(max_size=4)),
        max_size=4,
    ))
    link.patterns = draw(st.lists(
        st.tuples(st.text(min_size=1, max_size=3), attribute_dict(max_size=2)),
        max_size=3,
    ))
    link.features = set(draw(st.lists(st.text(max_size=4), max_size=4)))
    return link



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains unittests for vermouth.processors.annotate_mut_mod.
"""

import networkx as nx
import pytest
from vermouth.molecule import Molecule
from vermouth.forcefield import ForceField
from vermouth.processors.annotate_mut_mod import (
    parse_residue_spec,
    _subdict,
    _terminal_matches,
    annotate_modifications,
    AnnotateMutMod
)
from vermouth.tests.datafiles import (
    FF_UNIVERSAL_TEST,
)

# pylint: disable=redefined-outer-name

@pytest.fixture
def example_mol():
    mol = Molecule(force_field=ForceField(FF_UNIVERSAL_TEST))
    nodes = [
        {'chain': 'A', 'resname': 'A', 'resid': 1},  # 0, R1
        {'chain': 'A', 'resname': 'A', 'resid': 2},  # 1, R2
        {'chain': 'A', 'resname': 'A', 'resid': 2},  # 2, R2
        {'chain': 'A', 'resname': 'B', 'resid': 2},  # 3, R3
        {'chain': 'B', 'resname': 'A', 'resid': 1},  # 4, R4
        {'chain': 'B', 'resname': 'A', 'resid': 2},  # 5, R5
        {'chain': 'B', 'resname': 'A', 'resid': 2},  # 6, R5
        {'chain': 'B', 'resname': 'B', 'resid': 2},  # 7, R6
        {'chain': 'A', 'resname': 'C', 'resid': 3},  # 8, R7
    ]
    mol.add_nodes_from(enumerate(nodes))
    mol.add_edges_from([(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 6), (4, 7), (3, 8)])
    return mol


@pytest.mark.parametrize('spec,expected', [
    ('', {}),
    ('-', {'chain': ''}),
    ('#', {}),
    ('-#', {'chain': ''}),
    ('A-ALA1', {'chain': 'A', 'resname': 'ALA', 'resid': 1}),
    ('A-ALA#1', {'chain': 'A', 'resname': 'ALA', 'resid': 1}),
    ('ALA1', {'resname': 'ALA', 'resid': 1}),
    ('A-ALA', {'chain': 'A', 'resname': 'ALA'}),
    ('ALA', {'resname': 'ALA'}),
    ('2', {'resid': 2}),
    ('#2', {'resid': 2}),
    ('PO4#3', {'resname': 'PO4', 'resid': 3}),
    ('PO43', {'resname': 'PO', 'resid': 43}),
    ('A-B-C#D#1', {'chain': 'A', 'resid': 1, 'resname': 'B-C#D'}),
    ('ter', {'resname': 'ter'})
])
def test_parse_residue_spec(spec, expected):
    found = parse_residue_spec(spec)
    assert found == expected


@pytest.mark.parametrize('dict1,dict2,expected', [
    ({}, {}, True),
    ({1: 1}, {}, False),
    ({}, {1: 1}, True),
    ({1: 1}, {1: 1}, True),
    ({1: 1}, {1: 1, 2: 2}, True),
    ({1: 1, 2: 2}, {1: 1}, False),
    ({1: 1}, {1: 3, 2: 2}, False),
    ({1: 1, 2: 2}, {1: 1, 2: 3}, False),
])
def test_subdict(dict1, dict2, expected):
    found = _subdict(dict1, dict2)
    assert found == expected


@pytest.mark.parametrize('mutations,expected_mut', [
    ([], {}),
    ([({'chain': 'A', 'resname': 'A', 'resid': 1}, 'ALA')], {0: {'mutation': ['ALA']}}),
    (
        [({'resname': 'A'}, 'ALA')],
        {0: {'mutation': ['ALA']},
         1: {'mutation': ['ALA']},
         2: {'mutation': ['ALA']},
         4: {'mutation': ['ALA']},
         5: {'mutation': ['ALA']},
         6: {'mutation': ['ALA']},}
    ),
    (
        [({'resid': 2, 'chain': 'B'}, 'ALA')],
        {5: {'mutation': ['ALA']},
         6: {'mutation': ['ALA']},
         7: {'mutation': ['ALA']},}
    ),
    (
        [({'resname': 'cter'}, 'ALA')],
        {5: {'mutation': ['ALA']},
         6: {'mutation': ['ALA']},
         7: {'mutation': ['ALA']},
         8: {'mutation': ['ALA']}},
    ),
    (
        [({'resid': 2, 'chain': 'B'}, 'none')],  # none is not an existing modification...
        {5: {'mutation': ['none']},
         6: {'mutation': ['none']},
         7: {'mutation': ['none']},}
    ),
])
@pytest.mark.parametrize('modifications,expected_mod', [
    ([], {}),
    ([({'chain': 'A', 'resname': 'A', 'resid': 1}, 'C-ter')], {0: {'modification': ['C-ter']}}),
    (
        [({'chain': 'A', 'resname': 'A', 'resid': 1}, 'C-ter'),
         ({'chain': 'A', 'resname': 'A', 'resid': 1}, 'HSD')],
        {0: {'modification': ['C-ter', 'HSD']}}
    ),
    ([({'resname': 'B', 'resid': 1}, 'C-ter'),], {}),
    (
        [({'resname': 'B', 'resid': 2}, 'C-ter'),],
        {3: {'modification': ['C-ter']},
         7: {'modification': ['C-ter']},}
    ),
    ([({'resname': 'B', 'resid': 1}, 'C-ter'),], {}),
    (
        [({'resname': 'nter'}, 'C-ter')],
        {0: {'modification': ['C-ter']},},
    ),
    (
        [({'resname': 'cter', 'chain': 'B'}, 'C-ter'), ({'resname': 'C'}, 'HSD')],
        {7: {'modification': ['C-ter']},
         5: {'modification': ['C-ter']},
         6: {'modification': ['C-ter']},
         8: {'modification': ['HSD']}}  # Not a C-ter mod
    )
])
def test_annotate_modifications(example_mol, modifications, mutations, expected_mod, expected_mut):
    annotate_modifications(example_mol, modifications, mutations)
    for node_idx, mods in expected_mod.items():
        assert _subdict(mods, example_mol.nodes[node_idx])
    for node_idx, mods in expected_mut.items():
        assert _subdict(mods, example_mol.nodes[node_idx])


def test_single_residue_mol():
    mol = Molecule(force_field=ForceField(FF_UNIVERSAL_TEST))
    nodes = [
        {'chain': 'A', 'resname': 'A', 'resid': 2},
        {'chain': 'A', 'resname': 'A', 'resid': 2},
    ]
    mol.add_nodes_from(enumerate(nodes))
    mol.add_edges_from([(0, 1)])

    modification = [({'resname': 'A', 'resid': 2}, 'C-ter'),]
    annotate_modifications(mol, modification, [])

    assert mol.nodes[0] == {'modification': ['C-ter'], 'resname': 'A', 'resid': 2, 'chain': 'A'}
    assert mol.nodes[1] == {'modification': ['C-ter'], 'resname': 'A', 'resid': 2, 'chain': 'A'}


@pytest.mark.parametrize('modifications,mutations', [
    ([({'chain': 'A'}, 'M')], []),  # unknown residue name
    ([], [({'resid': 1}, 'M')]),  # unknown modification name
])
def test_annotate_modifications_error(example_mol, modifications, mutations):
    with pytest.raises(NameError):
        annotate_modifications(example_mol, modifications, mutations)


def test_unknown_terminus_match():
    resname = 'xter'
    graph = nx.Graph()
    graph.add_edge(0, 1)
    with pytest.raises(KeyError):
        _terminal_matches(resname, graph, 0)


@pytest.mark.parametrize('mutations,expected_mut', [
    ([], {}),
    ([('A-A1', 'ALA')], {0: {'mutation': ['ALA']}}),
    (
        [('A', 'ALA')],
        {0: {'mutation': ['ALA']},
         1: {'mutation': ['ALA']},
         2: {'mutation': ['ALA']},
         4: {'mutation': ['ALA']},
         5: {'mutation': ['ALA']},
         6: {'mutation': ['ALA']},}
    ),
    (
        [('B-2', 'ALA')],
        {5: {'mutation': ['ALA']},
         6: {'mutation': ['ALA']},
         7: {'mutation': ['ALA']},}
    )
])
@pytest.mark.parametrize('modifications,expected_mod', [
    ([], {}),
    ([('A-A1', 'C-ter')], {0: {'modification': ['C-ter']}}),
    (
        [('A-A1', 'C-ter'),
         ('A-A1', 'HSD')],
        {0: {'modification': ['C-ter', 'HSD']}}
    ),
    ([('B1', 'C-ter'),], {}),
    (
        [('B2', 'C-ter'),],
        {3: {'modification': ['C-ter']},
         7: {'modification': ['C-ter']},}
    ),
    ([('B1', 'C-ter'),], {}),
])
def test_annotate_mutmod_processor(example_mol, modifications, mutations, expected_mod, expected_mut):

    AnnotateMutMod(modifications, mutations).run_molecule(example_mol)
    for node_idx, mods in expected_mod.items():
        assert _subdict(mods, example_mol.nodes[node_idx])
    for node_idx, mods in expected_mut.items():
        assert _subdict(mods, example_mol.nodes[node_idx])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides Processors, VerMoUTH's work horses.
"""


from .gro_reader import GROInput
from .make_bonds import MakeBonds
from .pdb_reader import PDBInput
from .repair_graph import RepairGraph
from .do_mapping import DoMapping
from .do_links import DoLinks
from .average_beads import DoAverageBead
from .apply_posres import ApplyPosres
from .set_molecule_meta import SetMoleculeMeta
from .locate_charge_dummies import LocateChargeDummies
from .attach_mass import AttachMass
from .apply_rubber_band import ApplyRubberBand
from .merge_chains import MergeChains
from .canonicalize_modifications import CanonicalizeModifications
from .rename_modified_residues import RenameModifiedResidues
from .tune_cystein_bridges import (
    RemoveCysteinBridgeEdges,
    AddCysteinBridgesThreshold,
)
from .add_molecule_edges import AddMoleculeEdgesAtDistance, MergeNucleicStrands
from .name_moltype import NameMolType
from .quote import Quoter
from .go_vs_includes import GoVirtIncludes
from .sort_molecule_atoms import SortMoleculeAtoms
from .merge_all_molecules import MergeAllMolecules
from .annotate_mut_mod import AnnotateMutMod



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

try:
    import pkg_resources
except ImportError:
    import os
    TEST_DATA = os.path.join(os.path.dirname(__file__), 'data')
    del os
else:
    TEST_DATA = Path(pkg_resources.resource_filename('vermouth.tests', 'data'))
    del pkg_resources

# PDB files with a single molecule
PDB_PROTEIN = TEST_DATA / '1bta.pdb'
PDB_PARTIALLY_PROTEIN = TEST_DATA / '1bta_mutated.pdb'  # LYS replaced by UNK
PDB_NOT_PROTEIN = TEST_DATA / 'heme.pdb'
PDB_ALA5 = TEST_DATA / 'ala5.pdb'
PDB_ALA5_CG = TEST_DATA / 'ala5_cg.pdb'

# Full PDB files
PDB_CYS = TEST_DATA / '2QWO.pdb'  # Contains cystein bridges
PDB_HB = TEST_DATA / '2dn2.pdb'  # Hemoglobin with heme and 2*2 protein chains
PDB_MULTIMODEL = TEST_DATA / '6E8W.pdb'  # HIV-1, 15 models

# DNA
SHORT_DNA = TEST_DATA / 'dna-short.pdb'

DSSP_OUTPUT = TEST_DATA / 'dssp_1bta.ssd'

# Test force fields
FF_UNIVERSAL_TEST = TEST_DATA / 'force_fields' / 'universal-test'
FF_PEPPLANE = TEST_DATA / 'force_fields' / 'pepplane'
FF_MARTINI_TEST = TEST_DATA / 'force_fields' / 'martini-test'

# Mappings
MAP_UNIVERSAL_TEST_PEPPLANE = TEST_DATA / 'mappings' / 'universal-test'


# Clean the namespace so only the data file variables can be imported.
# An other option would be to define __all__, but it is easy to forget to add
# a variable in that list.
del Path



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides helper function for selecting part of a system, e.g. all proteins, or
protein backbones.
"""

import numpy as np
from .molecule import attributes_match


# TODO: Make that list part of the force fields
PROTEIN_RESIDUES = set(('ALA', 'ARG', 'ASP', 'ASN', 'CYS',
                        'GLU', 'GLN', 'GLY', 'HIS', 'ILE',
                        'LEU', 'LYS', 'MET', 'PHE', 'PRO',
                        'SER', 'THR', 'TRP', 'TYR', 'VAL'))


def is_protein(molecule):
    """
    Return True if all the residues in the molecule are protein residues.

    The function tests if the residue name of all the atoms in the input
    molecule are in ``PROTEIN_RESIDUES``.

    Parameters
    ----------
    molecule: Molecule
        The molecule to test.

    Returns
    -------
    bool
    """
    return all(
        molecule.nodes[n_idx].get('resname') in PROTEIN_RESIDUES
        for n_idx in molecule
    )


def select_all(_):
    """
    Returns True for all particles.
    """
    return True


# TODO: Have the backbone definition be force field specific.
def select_backbone(node):
    """
    Returns True if `node` is in a protein backbone.
    """
    return node.get('atomname') == 'BB'


def selector_has_position(atom):
    """
    Return True if the atom have a position.

    An atom is considered as not having a position if:
    * the "position" key is not defined;
    * the value of "position" is ``None``;
    * the coordinates are not finite numbers.

    Parameters
    ----------
    atom: dict

    Returns
    -------
    bool
    """
    position = atom.get('position')
    return position is not None and np.all(np.isfinite(position))


def proto_select_attribute_in(node, attribute, values):
    """
    Return True if the given attribute of the node is in a list of values.

    To be used as a selector, the function must be wrapped in a way that it can
    be called without the need to explicitly specify the 'attribute' and
    'values' arguments. This can be done using :func:`functools.partial`:

    >>> # select an atom if its name is in a given list
    >>> to_keep = ['BB', 'SC1']
    >>> select_name_in = functools.partial(
    ...     proto_select_attribute_in,
    ...     attribute='atomname',
    ...     values=to_keep
    ... )
    >>> select_name_in(node)

    Parameters
    ----------
    node: dict
        The atom/node to consider.
    attribute: str
        The key to look at in the node.
    values: list
        The values the node attribute can take for the node to be selected.

    Returns
    -------
    bool
    """
    return node.get(attribute) in values


def proto_multi_templates(node, templates, ignore_keys=()):
    """
    Return `True` is the node matched one of the templates.

    Parameters
    ----------
    node: dict
        The atom/node to consider.
    templates: collections.abc.Iterable[dict]
        A list of node templates to compare to the node.
    ignore_keys: collections.abc.Collection
        List of keys to ignore from the templates.

    Returns
    -------
    bool

    See Also
    --------
    vermouth.molecule.attributes_match
    """
    return any(
        attributes_match(node, template, ignore_keys)
        for template in templates
    )


def filter_minimal(molecule, selector):
    """
    Yield the atom keys that match the selector.

    The selector must be a function that accepts an atom as a argument. The
    atom is passed as a node attribute dictionary. The selector must return
    ``True`` for atoms to keep in the selection.

    The function can be used to build a subgraph that only contains the
    selection:

    .. code-block:: python

        selection = molecule.subgraph(
             filter_minimal(molecule, selector_function)
        )

    Parameters
    ----------
    molecule: Molecule
    selector: collections.abc.Callable

    Yields
    ------
    keys:
        Keys of the atoms that match the selection.
    """
    for name, atom in molecule.nodes.items():
        if selector(atom):
            yield name



# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Geometric operations.
"""

import numpy as np


def distance_matrix(coordinates_a, coordinates_b):
    """
    Compute a distance matrix between two set of points.

    Notes
    -----
    This function does **not** account for periodic boundary conditions.

    Parameters
    ----------
    coordinates_a: numpy.ndarray
        Coordinates of the points in the selections. Each row must correspond
        to a point and each column to a dimension.
    coordinates_b: numpy.ndarray
        Coordinates of the points in the selections. Each row must correspond
        to a point and each column to a dimension.

    Returns
    -------
    numpy.ndarray
        Rows correspond to the points from `coordinates_a`, columns correspond
        from `coordinates_b`.
    """
    return np.sqrt(
        np.sum(
            (coordinates_a[:, np.newaxis, :] - coordinates_b[np.newaxis, :, :]) ** 2,
            axis=-1)
    )


def angle(vector_ba, vector_bc):
    """
    Calculate the angle in radians between two vectors.

    The function assumes the following situation::

          B
         / \\
        A   C

    It returns the angle between BA and BC.
    """
    nominator = np.dot(vector_ba, vector_bc)
    denominator = np.linalg.norm(vector_ba) * np.linalg.norm(vector_bc)
    cosine = nominator / denominator
    # Floating errors at the limits may cause issues.
    cosine = np.clip(cosine, -1, 1)
    return np.arccos(cosine)


def dihedral(coordinates):
    """
    Calculate the dihedral angle in radians.

    Parameters
    ----------
    coordinates: numpy.ndarray
        The coordinates of 4 points defining the dihedral angle. Each row
        corresponds to a point, and each column to a dimension.

    Returns
    -------
    float
        The calculated angle between -pi and +pi.
    """
    vector_ab = coordinates[1, :] - coordinates[0, :]
    vector_bc = coordinates[2, :] - coordinates[1, :]
    vector_cd = coordinates[3, :] - coordinates[2, :]
    normal_abc = np.cross(vector_ab, vector_bc)
    normal_bcd = np.cross(vector_bc, vector_cd)
    psin = np.dot(normal_abc, vector_cd) * np.linalg.norm(vector_bc)
    pcos = np.dot(normal_abc, normal_bcd)
    return np.arctan2(psin, pcos)


def dihedral_phase(coordinates):
    """
    Calculate a dihedral angle in radians with a -pi phase correction.

    Parameters
    ----------
    coordinates: numpy.ndarray
        The coordinates of 4 points defining the dihedral angle. Each row
        corresponds to a point, and each column to a dimension.

    Returns
    -------
    float
        The calculated angle between -pi and +pi.

    See Also
    --------
    dihedral
        Calculate a dihedral angle.
    """
    dihedral_angle = dihedral(coordinates)
    dihedral_angle -= np.pi
    if dihedral_angle > np.pi:
        dihedral_angle -= 2 * np.pi
    if dihedral_angle < -np.pi:
        dihedral_angle += 2 * np.pi
    return dihedral_angle



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the :mod:`vermouth.processors.tune_cystein_bridges` module.
"""

# The redefined-outer-name check from pylint wrongly catches the use of pytest
# fixtures.
# pylint: disable=redefined-outer-name

import copy

import pytest

import vermouth
from vermouth.pdb.pdb import read_pdb
from vermouth.processors import tune_cystein_bridges
from vermouth.utils import distance
from vermouth.molecule import Choice
from .test_edge_tuning import simple_protein  # pylint: disable=unused-import
from .datafiles import PDB_CYS


@pytest.fixture
def cys_protein():
    """
    Read a PDB file describing a protein with a cystein bridge and cystein
    residues that are not involved in a bridge. Embed the content of the PDB in
    a :class:`vermouth.system.System`.

    Note that the PDB file does contain water, ions, and ligands in addition to
    the protein. The ligand bonds are explicitely provided as CONECT records.
    The edges from the CONECT records are removed from the system, so there is
    no edge left.
    """
    molecules = read_pdb(PDB_CYS)
    assert len(molecules) == 1
    molecule = molecules[0]
    molecule.remove_edges_from(list(molecule.edges))
    system = vermouth.system.System()
    system.molecules = [molecule]
    assert len(system.molecules[0].edges) == 0
    return system


@pytest.fixture
def simple_protein_pruned(simple_protein):
    """
    Protein-like molecule with cystein bridges removed using
    :func:`tune_cystein_bridges.remove_cystein_bridge_edges`.
    """
    graph = copy.deepcopy(simple_protein)
    tune_cystein_bridges.remove_cystein_bridge_edges(graph)
    return graph


@pytest.mark.parametrize('edge', ((5, 13), (7, 11)))
def test_remove_cystein_bridge_edges_remove(simple_protein_pruned, edge):
    """
    Assure that the expected edges were removed by
    :func:`tune_cystein_bridges.remove_cystein_bridge_edges`.
    """
    assert edge not in simple_protein_pruned.edges


@pytest.mark.parametrize('edge', (
    (0, 1), (2, 3), (2, 4), (4, 5), (4, 6), (6, 7), (6, 8), (8, 9),
    (8, 10), (10, 11), (10, 12), (12, 13), (12, 14), (14, 15),
    (14, 16), (16, 17), (1, 17), (3, 15),
))
def test_remove_cystein_bridge_edges_kept(simple_protein_pruned, edge):
    """
    Assure that :func:`tune_cystein_bridges.remove_cystein_bridge_edges`
    does not remove edges it should not remove.
    """
    assert edge in simple_protein_pruned.edges


def test_add_cystein_bridges_threshold(cys_protein):
    """
    Test that :class:`vermouth.AddCysteinBridgesThreshold` detects a cystein
    bridge in a PDB file.
    """
    processor = vermouth.AddCysteinBridgesThreshold(threshold=0.22)
    system = processor.run_system(cys_protein)
    for edge in system.molecules[0].edges:
        node_a = system.molecules[0].nodes[edge[0]]
        node_b = system.molecules[0].nodes[edge[1]]
        print('*', node_a, '--', node_b)
        print(distance(node_a['position'], node_b['position']))
    assert len(system.molecules[0].edges) == 1
    # According to SSBOND record, cysbond should be between resid 171 and 876.
    # Let's find the associated atoms.
    atoms = tuple(system.molecules[0].find_atoms(resname='CYS', atomname='SG',
                                                 resid=Choice([171, 876])))
    assert atoms in system.molecules[0].edges


def test_remove_cystein_bridge_edges_processor(cys_protein):
    """
    Test that :class:`vermouth.RemoveCysteinBridgeEdges` removes edges.
    """
    print(cys_protein.molecules[0].nodes[1285], cys_protein.molecules[0].nodes[3508])
    cys_protein.molecules[0].add_edge(1285, 3508)
    processor = vermouth.RemoveCysteinBridgeEdges()
    processor.run_system(cys_protein)
    # According to SSBOND record, cysbond should be between resid 171 and 876.
    # Let's find the associated atoms.
    atoms = tuple(cys_protein.molecules[0].find_atoms(resname='CYS',
                                                      atomname='SG',
                                                      resid=Choice([171, 876])))
    assert atoms not in cys_protein.molecules[0].edges



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import itertools

import hypothesis
import hypothesis.strategies as st
import numpy as np
import pytest
import vermouth
import vermouth.molecule
import vermouth.forcefield
from vermouth.molecule import Interaction, Molecule

from .molecule_strategies import random_molecule, random_block, random_link

# pylint: disable=redefined-outer-name, no-value-for-parameter


# In some cases, drawing from hypothesis.strategies.text is extremely slow and
# triggers the heath check warnings. It appears to be due to some internal
# combinations of timers and cache. This is a workwround that forces the
# generation, and the caching, of the character table out of any test. This
# way, the table is cached, and the time required to build it is not accounted
# in the health checks.
# See <https://github.com/HypothesisWorks/hypothesis/issues/1153>.
hypothesis.internal.charmap.charmap()


@pytest.fixture
def molecule():
    molecule = vermouth.molecule.Molecule()
    molecule.meta['test'] = True
    molecule.meta['test_mutable'] = [0, 1, 2]
    # The node keys should not be in a sorted order as it would mask any issue
    # due to the keys being accidentally sorted.
    molecule.add_node(2, atomname='CC')
    molecule.add_node(0, atomname='AA', mutable=[7, 8, 9])
    molecule.add_node(1, atomname='BB')
    molecule.add_edge(0, 1)
    molecule.add_edge(0, 2)
    molecule.add_interaction(
        type_='bonds',
        atoms=(0, 1),
        parameters=['1', '2'],
        meta={'unmutable': 0, 'mutable': [4, 5, 6]},
    )
    molecule.add_interaction(
        type_='bonds',
        atoms=(0, 2),
        parameters=['a', 'b'],
    )
    return molecule


@pytest.fixture
def molecule_copy(molecule):
    return molecule.copy()


@pytest.mark.parametrize('atoms, bonds, interactions, removed, expected', [
    # empty molecule
    ([], [], [], [], {}),
    # Nodes but no interactions
    ([1, 2, 3], [(1, 2), (2, 3)], [], [2], {}),
    # interactions that all need to be removed
    ([1, 2], [(1, 2)], [('bond', (1, 2), {})], [2], {}),
    # Molecule with interactions of which some need to be removed
    (
        [1, 2, 3, 4, 5, 6, 7],
        [(1, 2), (3, 4), (5, 7)],
        [('bond', (5, 4), {}), ('bond', (1, 2), {})],
        [5],
        {'bond': [vermouth.molecule.Interaction(atoms=(1, 2), meta={}, parameters={})]}
    ),
    # Molecule with interactions of which none need to be removed
    (
        [1, 2, 3, 4, 5, 6, 7],
        [(1, 2), (3, 4)],
        [('bond', (5, 4), {}), ('bond', (1, 2), {})],
        [6, 7],
        {'bond': [vermouth.molecule.Interaction(atoms=(5, 4), meta={}, parameters={}),
                  vermouth.molecule.Interaction(atoms=(1, 2), meta={}, parameters={})]}
    ),
    # Molecule with interactions of different types of which all need to be removed
    (
        [1, 2, 3, 4, 5, 6, 7],
        [(1, 2), (3, 4), (5, 7), (6, 7)],
        [('bond', (1, 6), {}), ('bond', (5, 4), {}), ('angle', (1, 2), {})],
        [1, 5],
        {}
    ),
    # Molecule with interactions of different types of which some need to be removed
    (
        [1, 2, 3, 4, 5, 6, 7],
        [(1, 2), (3, 4), (5, 7), (6, 7)],
        [('bond', (1, 6), {}), ('bond', (5, 4), {}), ('angle', (1, 2), {})],
        [5],
        {'angle': [vermouth.molecule.Interaction(atoms=(1, 2), meta={}, parameters={})],
         'bond': [vermouth.molecule.Interaction(atoms=(1, 6), meta={}, parameters={})]}
    ),
    # Molecule with interactions of different types of which none need to be removed
    (
        [1, 2, 3, 4, 5, 6, 7],
        [(1, 2), (3, 4), (5, 7), (6, 7)],
        [('bond', (1, 6), {}), ('bond', (5, 4), {}), ('angle', (1, 2), {})],
        [3],
        {'angle': [vermouth.molecule.Interaction(atoms=(1, 2), meta={}, parameters={})],
         'bond': [vermouth.molecule.Interaction(atoms=(1, 6), meta={}, parameters={}),
                  vermouth.molecule.Interaction(atoms=(5, 4), meta={}, parameters={})]}
    ),
])
def test_remove_nodes_from(atoms, bonds, interactions, removed, expected):
    """
    Test whether molecule.remove_nodes_from also removes the corresponding
    interactions
    """
    molecule = vermouth.molecule.Molecule()
    molecule.add_nodes_from(atoms)
    molecule.add_edges_from(bonds)
    for type_, atoms, params in interactions:
        molecule.add_interaction(type_, atoms, params)

    molecule.remove_nodes_from(removed)
    assert molecule.interactions == expected


@pytest.mark.parametrize('atoms, bonds, interactions, removed, expected', [
    # empty molecule
    ([], [], [], None, {}),
    # Nodes but no interactions
    ([1, 2, 3], [(1, 2), (2, 3)], [], 2, {}),
    # interactions that all need to be removed
    ([1, 2], [(1, 2)], [('bond', (1, 2), {})], 2, {}),
    # Molecule with interactions of which some need to be removed
    (
        [1, 2, 3, 4, 5, 6, 7],
        [(1, 2), (3, 4), (5, 7)],
        [('bond', (5, 4), {}), ('bond', (1, 2), {})],
        5,
        {'bond': [vermouth.molecule.Interaction(atoms=(1, 2), meta={}, parameters={})]}
    ),
    # Molecule with interactions of different types of which some need to be removed
    (
        [1, 2, 3, 4, 5, 6, 7],
        [(1, 2), (3, 4), (5, 7), (6, 7)],
        [('bond', (1, 6), {}), ('bond', (5, 4), {}), ('angle', (1, 2), {})],
        5,
        {'angle': [vermouth.molecule.Interaction(atoms=(1, 2), meta={}, parameters={})],
         'bond': [vermouth.molecule.Interaction(atoms=(1, 6), meta={}, parameters={})]}
    ),
    # Molecule with interactions of different types of which none need to be removed
    (
        [1, 2, 3, 4, 5, 6, 7],
        [(1, 2), (3, 4), (5, 7), (6, 7)],
        [('bond', (1, 6), {}), ('bond', (5, 4), {}), ('angle', (1, 2), {})],
        3,
        {'angle': [vermouth.molecule.Interaction(atoms=(1, 2), meta={}, parameters={})],
         'bond': [vermouth.molecule.Interaction(atoms=(1, 6), meta={}, parameters={}),
                  vermouth.molecule.Interaction(atoms=(5, 4), meta={}, parameters={})]}
    ),
    (
        [1, 2, 3, 4],
        [(1, 2), (2, 3), (3, 4)],
        [('bond', (1, 2), {}), ('bond', (2, 3), {}), ('bond', (3, 4), {})],
        2,
        {'bond': [vermouth.molecule.Interaction(atoms=(3, 4), meta={}, parameters={})]},
    ),
])
def test_remove_node(atoms, bonds, interactions, removed, expected):
    """
    Test whether molecule.remove_node also removes the corresponding
    interactions
    """
    molecule = vermouth.molecule.Molecule()
    molecule.add_nodes_from(atoms)
    molecule.add_edges_from(bonds)
    for type_, atoms, params in interactions:
        molecule.add_interaction(type_, atoms, params)

    if removed:
        molecule.remove_node(removed)

    assert molecule.interactions == expected

@pytest.fixture
def molecule_subgraph(molecule):
    return molecule.subgraph([2, 0])


def test_copy(molecule, molecule_copy):
    assert molecule_copy is not molecule
    assert molecule_copy.meta == molecule.meta
    assert list(molecule_copy.nodes) == list(molecule.nodes)
    assert list(molecule_copy.nodes.values()) == list(molecule.nodes.values())
    assert molecule_copy.interactions == molecule.interactions


def test_copy_meta_mod(molecule, molecule_copy):
    molecule_copy.meta['test'] = False
    assert molecule_copy.meta['test'] != molecule.meta['test']
    # We are doing a copy, not a deep copy.
    assert molecule_copy.meta['test_mutable'] is molecule.meta['test_mutable']


def test_copy_node_mod(molecule, molecule_copy):
    molecule_copy.nodes[0]['atomname'] = 'mod'
    assert molecule_copy.nodes[0]['atomname'] != molecule.nodes[0]['atomname']
    extra_value = 'a new attribute'
    molecule_copy.nodes[0]['extra'] = extra_value
    assert molecule_copy.nodes[0]['extra'] == extra_value
    assert 'extra' not in molecule.nodes

    # We are looking at a copy, not a deep copy
    assert molecule_copy.nodes[0]['mutable'] is molecule.nodes[0]['mutable']

    molecule_copy.add_node('new')
    assert 'new' in molecule_copy.nodes
    assert 'new' not in molecule


def test_copy_edge_mod(molecule, molecule_copy):
    molecule_copy.add_edge(1, 2)
    assert (1, 2) in molecule_copy.edges
    assert (1, 2) not in molecule.edges
    molecule_copy.edges[(0, 1)]['attribute'] = 1
    assert molecule_copy.edges[(0, 1)]['attribute'] == 1
    assert 'attribute' not in molecule.edges[(0, 1)]


def test_copy_interactions_mod(molecule, molecule_copy):
    molecule_copy.add_interaction(
        type_='bonds',
        atoms=(0, 2),
        parameters=['3', '4'],
        meta={'unmutable': 0},
    )
    n_bonds = len(molecule.interactions['bonds'])
    n_bonds_copy = len(molecule_copy.interactions['bonds'])
    assert n_bonds_copy > n_bonds


def test_subgraph_base(molecule_subgraph):
    assert tuple(molecule_subgraph) == (2, 0)  # order matters!
    assert (0, 2) in molecule_subgraph.edges
    assert (0, 1) not in molecule_subgraph.edges  # node 1 is not there


def test_subgraph_interactions(molecule_subgraph):
    bond_atoms = [bond.atoms for bond in molecule_subgraph.interactions['bonds']]
    assert (0, 2) in bond_atoms
    assert (0, 1) not in bond_atoms


def test_link_predicate_match():
    lp = vermouth.molecule.LinkPredicate(None)
    with pytest.raises(NotImplementedError):
        lp.match(1, 2)


@pytest.fixture
def edges_between_molecule():
    """
    Build an empty molecule with known connectivity.

    The molecule does not have any node attribute nor any molecule metadata. It
    only has a bare graph with a few nodes and edges.

    The graph looks like::

        0 - 1 - 3 - 4 - 5 - 7 - 8     9 - 10 - 11 - 12
            |           |
            2           6

    """
    molecule = vermouth.molecule.Molecule()
    molecule.add_edges_from((
        (0, 1), (1, 2), (1, 3), (3, 4), (4, 5), (5, 6), (5, 7), (7, 8),
        (9, 10), (10, 11), (11, 12),
    ))
    for node1, node2, attributes in molecule.edges(data=True):
        attributes['arbitrary'] = '{} - {}'.format(min(node1, node2),
                                                   max(node1, node2))
    return molecule


@pytest.fixture
def edges_between_selections():
    """
    Build a static list of selections of nodes from :func:`edges_between_molecule`.
    """
    return [
        (0, 1, 2, 3),
        (5, 6, 7, 8),
        (9, 10, 11, 12),
        (3, 4, 5, 6),
        (7, 8, 9, 10),
    ]


@pytest.mark.parametrize('data', [True, False])
@pytest.mark.parametrize('bunch1, bunch2, expected', (
    (0, 1, []), (0, 2, []), (0, 4, []), (1, 2, []), (2, 3, []),  # non-overlapping
    (0, 3, [(1, 3), (3, 4)]), (1, 3, [(4, 5), (5, 6), (5, 6), (5, 7)]),
    (1, 4, [(5, 7), (7, 8), (7, 8)]), (2, 4, [(9, 10), (9, 10), (10, 11)]),
))
def test_edges_between(edges_between_molecule, edges_between_selections,
                       bunch1, bunch2, expected, data):
    """
    Test :meth:`vermouth.molecule.Molecule.edges_between`.
    """
    selection_1 = edges_between_selections[bunch1]
    selection_2 = edges_between_selections[bunch2]
    found = list(edges_between_molecule.edges_between(
        selection_1, selection_2, data=data
    ))
    sorted_found = sorted(sorted(edge[:2]) for edge in found)
    sorted_expected = sorted(sorted(edge) for edge in expected)
    assert sorted_found == sorted_expected
    if data:
        found_attributes = [
            edge[2]
            for edge in sorted(found, key=lambda x: x[:2])
        ]
        expected_attributes = [
            edges_between_molecule.edges[edge[0], edge[1]]
            for edge in sorted_expected
        ]
        assert found_attributes == expected_attributes



@pytest.mark.parametrize('selidx, expected', (
    (0, ((0, 1), (1, 2), (1, 3))),
    (1, ((5, 6), (5, 7), (7, 8))),
    (2, ((9, 10), (10, 11), (11, 12))),
    (3, ((3, 4), (4, 5), (5, 6))),
    (4, ((7, 8), (9, 10))),
))
def test_subgraph_edges(edges_between_molecule, edges_between_selections,
                        selidx, expected):
    """
    :meth:`vermouth.molecule.Molecule.subgraph` select the expected edges.

    See Also
    --------
    test_subgraph_base
        This test deals with a larger graph, but no metadata; while the graph
        in :func:`test_subgraph_base` uses a very small and simple selection
        but the graph has metadata.
    """
    subgraph = edges_between_molecule.subgraph(edges_between_selections[selidx])
    sorted_found = sorted(sorted(edge) for edge in subgraph.edges)
    sorted_expected = sorted(sorted(edge) for edge in expected)
    assert sorted_found == sorted_expected

    found_attributes = [
        subgraph.edges[edge[0], edge[1]]
        for edge in sorted_expected
    ]
    expected_attributes = [
        edges_between_molecule.edges[edge[0], edge[1]]
        for edge in sorted_expected
    ]
    assert found_attributes == expected_attributes


@pytest.mark.parametrize('left, right, expected', (
    (  # Same
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=['a', '0.2', '200'],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['1', '0.2', '200'],
                            meta={'a': 0}),
            ],
        },
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=['a', '0.2', '200'],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['1', '0.2', '200'],
                            meta={'a': 0}),
            ],
        },
        True,
    ),
    (  # Difference in atoms
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=['a', '0.2', '200'],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['1', '0.2', '200'],
                            meta={'a': 0}),
            ],
        },
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=['a', '0.2', '200'],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'notC'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['1', '0.2', '200'],
                            meta={'a': 0}),
            ],
        },
        False,
    ),
    (  # Difference in parameters
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=['a', '0.2', '200'],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['1', '0.2', '200'],
                            meta={'a': 0}),
            ],
        },
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=['a', '0.2', '200'],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['different'],
                            meta={'a': 0}),
            ],
        },
        False,
    ),
    (  # Difference in meta
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=['a', '0.2', '200'],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['1', '0.2', '200'],
                            meta={'a': 0}),
            ],
        },
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=['a', '0.2', '200'],
                            meta={'a': 0, 'other': True}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['1', '0.2', '200'],
                            meta={'a': 0}),
            ],
        },
        False,
    ),
    (  # Equal with LinkParameterEffector
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=[
                                'a',
                                vermouth.molecule.ParamDistance(['A', 'B']),
                                '200',
                            ],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['1', '0.2', '200'],
                            meta={'a': 0}),
            ],
        },
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=[
                                'a',
                                vermouth.molecule.ParamDistance(['A', 'B']),
                                '200',
                            ],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['1', '0.2', '200'],
                            meta={'a': 0}),
            ],
        },
        True,
    ),
    (  # Different arguments for LinkParameterEffector
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=[
                                'a',
                                vermouth.molecule.ParamDistance(['A', 'B']),
                                '200',
                            ],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['1', '0.2', '200'],
                            meta={'a': 0}),
            ],
        },
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=[
                                'a',
                                vermouth.molecule.ParamDistance(['A', 'C']),
                                '200',
                            ],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['1', '0.2', '200'],
                            meta={'a': 0}),
            ],
        },
        False,
    ),
    (  # Different format_spec in LinkParameterEffector
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=[
                                'a',
                                vermouth.molecule.ParamDistance(['A', 'B']),
                                '200',
                            ],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['1', '0.2', '200'],
                            meta={'a': 0}),
            ],
        },
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=[
                                'a',
                                vermouth.molecule.ParamDistance(
                                    ['A', 'B'], format_spec='.2f',
                                ),
                                '200',
                            ],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['1', '0.2', '200'],
                            meta={'a': 0}),
            ],
        },
        False,
    ),
    (  # Different LinkParameterEffector
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=[
                                'a',
                                vermouth.molecule.ParamDistance(['A', 'B']),
                                '200',
                            ],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['1', '0.2', '200'],
                            meta={'a': 0}),
            ],
        },
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=[
                                'a',
                                vermouth.molecule.ParamAngle(['A', 'B', 'C']),
                                '200',
                            ],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['1', '0.2', '200'],
                            meta={'a': 0}),
            ],
        },
        False,
    ),
    (  # Missing interaction category ("angles" is missing in one molecule)
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=[
                                'a',
                                vermouth.molecule.ParamAngle(['A', 'B', 'C']),
                                '200',
                            ],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [
                Interaction(atoms=('A', 'B', 'C'),
                            parameters=['1', '0.2', '200'],
                            meta={'a': 0}),
            ],
        },
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=[
                                'a',
                                vermouth.molecule.ParamAngle(['A', 'B', 'C']),
                                '200',
                            ],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
        },
        False,
    ),
    (  # Empty interaction category
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=[
                                'a',
                                vermouth.molecule.ParamAngle(['A', 'B', 'C']),
                                '200',
                            ],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
            'angles': [],  # Empty!
        },
        {
            'bonds': [
                Interaction(atoms=('A', 'B'),
                            parameters=[
                                'a',
                                vermouth.molecule.ParamAngle(['A', 'B', 'C']),
                                '200',
                            ],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),
            ],
        },
        True,
    ),

))
def test_same_interactions(left, right, expected):
    """
    Test that Molecule.same_interactions works as expected.
    """
    left_mol = Molecule()
    left_mol.interactions = left
    right_mol = Molecule()
    right_mol.interactions = right
    assert left_mol.same_interactions(right_mol) == expected
    assert right_mol.same_interactions(left_mol) == expected


@pytest.mark.parametrize('left, right, expected', (
    (  # Simple identical
        (  # left
            (0, {'a': 'abc', 'b': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        (  # right
            (0, {'a': 'abc', 'b': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        True,  # expected
    ),
    (  # Wrong order
        (  # left
            (0, {'a': 'abc', 'b': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        (  # right
            (1, {'c': (0, 1, 2), 'd': None}),
            (0, {'a': 'abc', 'b': 123}),
        ),
        False,  # expected
    ),

    (  # Different string
        (  # left
            (0, {'a': 'abc', 'b': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        (  # right
            (0, {'a': 'different', 'b': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        False,
    ),
    (  # Different number
        (  # left
            (0, {'a': 'abc', 'b': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        (  # right
            (0, {'a': 'abc', 'b': 900}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        False,  # expected
    ),
    (  # Different tuple
        (  # left
            (0, {'a': 'abc', 'b': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        (  # right
            (0, {'a': 'abc', 'b': 123}),
            (1, {'c': (3, 2, 1), 'd': None}),
        ),
        False,  # expected
    ),
    (  # Equal Numpy array
        (  # left
            (0, {'a': np.linspace(2, 5, num=7), 'b': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        (  # right
            (0, {'a': np.linspace(2, 5, num=7), 'b': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        True,  # expected
    ),
    (  # Different Numpy array
        (  # left
            (0, {'a': np.linspace(2, 5, num=7), 'b': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        (  # right
            (0, {'a': np.linspace(2, 8, num=7), 'b': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        False,  # expected
    ),
    (  # Different shaped Numpy array
        (  # left
            (0, {'a': np.linspace(2, 5, num=9), 'b': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        (  # right
            (0, {'a': np.linspace(2, 5, num=9).reshape((3, 3)), 'b': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        False,  # expected
    ),
    (  # Mismatch types
        (  # left
            (0, {'a': np.linspace(2, 5, num=9), 'b': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        (  # right
            (0, {'a': 'not an array', 'b': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        False,  # expected
    ),
    (  # Mismatch attribute key
        (  # left
            (0, {'a': 'abc', 'b': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        (  # right
            (0, {'a': 'abc', 'different': 123}),
            (1, {'c': (0, 1, 2), 'd': None}),
        ),
        False,  # expected
    ),
    (  # same dicts as value
        (  # left
            (0, {'a': {'a': 0, 'b': 1}}),
        ),
        (  # right
            (0, {'a': {'a': 0, 'b': 1}}),
        ),
        True,
    ),
    (  # different dicts as value
        (  # left
            (0, {'a': {'a': 0, 'b': 1}}),
        ),
        (  # right
            (0, {'a': {'a': None, 'b': 1}}),
        ),
        False,
    ),
))
def test_same_nodes(left, right, expected):
    left_mol = Molecule()
    left_mol.add_nodes_from(left)
    right_mol = Molecule()
    right_mol.add_nodes_from(right)
    assert left_mol.same_nodes(right_mol) == expected
    assert right_mol.same_nodes(left_mol) == expected


@pytest.mark.parametrize('effector_class', (
    vermouth.molecule.ParamDistance,
    vermouth.molecule.ParamAngle,
    vermouth.molecule.ParamDihedral,
    vermouth.molecule.ParamDihedralPhase,
))
@pytest.mark.parametrize('format_spec', (
    None, '.2f', '0.3f',
))
def test_link_parameter_effector_equal(effector_class, format_spec):
    """
    Test that equal LinkParameterEffector compare equal.
    """
    n_keys = effector_class.n_keys_asked
    left_keys = ['A{}'.format(idx) for idx in range(n_keys)]
    right_keys = copy.copy(left_keys)  # Let's be sure the id is different
    left = effector_class(left_keys, format_spec=format_spec)
    right = effector_class(right_keys, format_spec=format_spec)
    assert left == right


@pytest.mark.parametrize('effector_class', (
    vermouth.molecule.ParamDistance,
    vermouth.molecule.ParamAngle,
    vermouth.molecule.ParamDihedral,
    vermouth.molecule.ParamDihedralPhase,
))
@pytest.mark.parametrize('format_right, format_left', itertools.combinations(
    (None, '.2f', '0.3f', ), 2
))
def test_link_parameter_effector_diff_format(effector_class, format_left, format_right):
    """
    Test that two instances of LinkParameterEffector compare different if they
    have different formats.
    """
    n_keys = effector_class.n_keys_asked
    left_keys = ['A{}'.format(idx) for idx in range(n_keys)]
    right_keys = copy.copy(left_keys)  # Let's be sure the id is different
    left = effector_class(left_keys, format_spec=format_left)
    right = effector_class(right_keys, format_spec=format_right)
    assert left != right


@pytest.mark.parametrize('effector_class', (
    vermouth.molecule.ParamDistance,
    vermouth.molecule.ParamAngle,
    vermouth.molecule.ParamDihedral,
    vermouth.molecule.ParamDihedralPhase,
))
def test_link_parameter_effector_diff_keys(effector_class):
    """
    Test that LinkParameterEffector compare different if they have different keys.
    """
    n_keys = effector_class.n_keys_asked
    left_keys = ['A{}'.format(idx) for idx in range(n_keys)]
    right_keys = ['B{}'.format(idx) for idx in range(n_keys)]
    left = effector_class(left_keys)
    right = effector_class(right_keys)
    assert left != right


@pytest.mark.parametrize('left_class, right_class', itertools.combinations((
    vermouth.molecule.ParamDistance,
    vermouth.molecule.ParamAngle,
    vermouth.molecule.ParamDihedral,
    vermouth.molecule.ParamDihedralPhase,
), 2))
def test_link_parameter_effector_diff_class(left_class, right_class):
    """
    Test that LinkParameterEffector compare different if they have different classes.
    """
    left_n_keys = left_class.n_keys_asked
    left_keys = ['A{}'.format(idx) for idx in range(left_n_keys)]
    left = left_class(left_keys)

    right_n_keys = right_class.n_keys_asked
    right_keys = ['A{}'.format(idx) for idx in range(right_n_keys)]
    right = right_class(right_keys)

    assert left != right


@hypothesis.settings(suppress_health_check=[hypothesis.HealthCheck.too_slow])
@hypothesis.given(random_molecule())
def test_molecule_equal(mol):
    """
    Two equal molecules are equal.
    """
    mol_copy = mol.copy()
    assert mol == mol_copy
    assert mol is not mol_copy


@hypothesis.given(random_block())
def test_block_equal(block):
    """
    Two equal blocks are equal.
    """
    block_copy = block.copy()
    assert block == block_copy
    assert block is not block_copy


@hypothesis.given(random_link(max_nodes=3, max_meta=3))
def test_link_equal(link):
    """
    Two equal links are equal.
    """
    link_copy = link.copy()
    assert link == link
    assert link is not link_copy


@pytest.mark.parametrize('left, right, expected', (
    (  # Equal edges, same order, same direction, same attributes
        [
            (0, 1, {'a': 0, 'b': [2, 3]}),
            (1, 3, {'foo': 'bar'}),
        ],
        [
            (0, 1, {'a': 0, 'b': [2, 3]}),
            (1, 3, {'foo': 'bar'}),
        ],
        True,
    ),
    (  # Equal edges, but different order
        [
            (0, 1, {'a': 0, 'b': [2, 3]}),
            (1, 3, {'foo': 'bar'}),
        ],
        [
            (1, 3, {'foo': 'bar'}),
            (0, 1, {'a': 0, 'b': [2, 3]}),
        ],
        True,
    ),
    (  # Equal edges, but different order and different direction
        [
            (0, 1, {'a': 0, 'b': [2, 3]}),
            (1, 3, {'foo': 'bar'}),
        ],
        [
            (3, 1, {'foo': 'bar'}),
            (0, 1, {'a': 0, 'b': [2, 3]}),
        ],
        True,
    ),
    (  # Edges only differ by an attribute
        [
            (0, 1, {'a': 0, 'b': [2, 3]}),
            (1, 3, {'foo': 'bar'}),
        ],
        [
            (0, 1, {'a': 0, 'b': [2, 40]}),
            (1, 3, {'foo': 'bar'}),
        ],
        False,
    ),
    (  # Edges are different
        [
            (0, 1, {'foo': 'bar'}),
            (2, 3, {}),
        ],
        [
            (5, 6, {'foo': 'bar'}),
            (3, 4, {}),
        ],
        False,
    ),
))
def test_same_edges(left, right, expected):
    """
    Compare edges between two molecules.
    """
    molecule_left = vermouth.molecule.Molecule()
    molecule_left.add_edges_from(left)
    molecule_right = vermouth.molecule.Molecule()
    molecule_right.add_edges_from(right)
    assert molecule_left.same_edges(molecule_right) == expected
    assert molecule_right.same_edges(molecule_left) == expected


@pytest.mark.parametrize('left, right, expected', (
    (  # Identical non-edges
        [
            ('XX', {'foo': 'bar', 'toto': [0, 1, 2]}),
            ('XX', {'foo': 'no bar', 'plop': 'else'}),
            ('YY', {'klonk': 0}),
        ],
        [
            ('XX', {'foo': 'bar', 'toto': [0, 1, 2]}),
            ('XX', {'foo': 'no bar', 'plop': 'else'}),
            ('YY', {'klonk': 0}),
        ],
        True,
    ),
    (  # Identical non-edges, but disordered
        [
            ('XX', {'foo': 'no bar', 'plop': 'else'}),
            ('YY', {'klonk': 0}),
            ('XX', {'foo': 'bar', 'toto': [0, 1, 2]}),
        ],
        [
            ('XX', {'foo': 'bar', 'toto': [0, 1, 2]}),
            ('XX', {'foo': 'no bar', 'plop': 'else'}),
            ('YY', {'klonk': 0}),
        ],
        True,
    ),

    (  # Different number of non-edges
        [
            ('XX', {'foo': 'bar', 'toto': [0, 1, 2]}),
            ('XX', {'foo': 'no bar', 'plop': 'else'}),
            ('YY', {'klonk': 0}),
        ],
        [
            ('XX', {'foo': 'bar', 'toto': [0, 1, 2]}),
            ('YY', {'klonk': 0}),
        ],
        False,
    ),
    (  # Same number of non-edges, but mismatched anchors
        [
            ('XX', {'foo': 'bar', 'toto': [0, 1, 2]}),
            ('XX', {'foo': 'no bar', 'plop': 'else'}),
            ('YY', {'klonk': 0}),
        ],
        [
            ('AA', {'foo': 'bar', 'toto': [0, 1, 2]}),
            ('BB', {'foo': 'no bar', 'plop': 'else'}),
            ('YY', {'klonk': 0}),
        ],
        False,
    ),
    (  # Mismatch in an attribute
        [
            ('XX', {'foo': 'bar', 'toto': [0, 1, 2]}),
            ('XX', {'foo': 'no bar', 'plop': 'else'}),
            ('YY', {'klonk': 0}),
        ],
        [
            ('XX', {'foo': 'mismatch', 'toto': [0, 1, 2]}),
            ('XX', {'foo': 'no bar', 'plop': 'else'}),
            ('YY', {'klonk': 0}),
        ],
        False,
    ),
    (  # Mismatch in number of attributes, there is a shortcut in same_non_edges
        [
            ('XX', {'foo': 'bar', 'toto': [0, 1, 2]}),
            ('XX', {'foo': 'no bar'}),
            ('YY', {}),
        ],
        [
            ('XX', {'foo': 'bar', 'toto': [0, 1, 2]}),
            ('XX', {'foo': 'no bar', 'plop': 'else'}),
            ('YY', {'klonk': 0}),
        ],
        False,
    ),
    (
        [
            ('XX', {'foo': 'bar', 'toto': [0, 1, 2]}),
            ('XX', {'foo': 'no bar'}),
            ('XX', {'foo': 'no bar again'}),
            ('YY', {}),
        ],
        [
            ('XX', {'foo': 'bar', 'toto': [0, 1, 2]}),
            ('XX', {'foo': 'no bar'}),
            ('YY', {}),
            ('YY', {'something': 'tadaam'}),
        ],
        False,
    ),
))
def test_same_non_edges(left, right, expected):
    """
    Compare non-edges between two links.
    """
    link_left = vermouth.molecule.Link()
    link_left.non_edges = left
    link_right = vermouth.molecule.Link()
    link_right.non_edges = right
    assert link_left.same_non_edges(link_right) == expected
    assert link_right.same_non_edges(link_left) == expected


@pytest.mark.parametrize('interactions, expected', (
    (
        {},
        []
    ),
    (
        {'bonds': []},
        []
    ),
    (
        {'bonds': [Interaction(atoms=(1, 2), parameters=[], meta=[])]},
        ['bonds']
    ),
    (
        {'bonds': [Interaction(atoms=(1, 2), parameters=[], meta=[])],
         'angles': []},
        ['bonds']
    ),
    (
        {'bonds': [Interaction(atoms=(1, 2), parameters=[], meta=[])],
         'angles': [Interaction(atoms=(1, 2, 3), parameters=[], meta=[])]},
        ['bonds', 'angles']
    ),
    (
        {'angles': [Interaction(atoms=(1, 2, 3), parameters=[], meta=[])],
         'bonds': [Interaction(atoms=(1, 2), parameters=[], meta=[])],},
        ['bonds', 'angles']
    ),
    (
        {'angles': [Interaction(atoms=(1, 2, 3), parameters=[], meta=[])],
         'a_bonds': [Interaction(atoms=(10, 2), parameters=[], meta=[])],
         'bonds': [Interaction(atoms=(1, 2), parameters=[], meta=[])],},
        ['a_bonds', 'bonds', 'angles']
    ),
))
def test_interaction_sort(interactions, expected):
    """
    Test the order produced by Molecule._sort_interactions.
    """
    assert vermouth.molecule.Molecule.sort_interactions(interactions) == expected


@hypothesis.given(moltype=st.one_of(st.none(), st.text()), mol=random_molecule())
def test_str_method(mol, moltype):
    """
    Test Molecule.__str__
    """
    if moltype is not None:
        mol.meta['moltype'] = moltype
    else:
        moltype = 'molecule'
    found = str(mol)
    assert '{} with '.format(moltype) in found
    assert '{} {}'.format(len(mol), 'atoms') in found
    hypothesis.assume('' not in mol.interactions)
    for itype, interactions in mol.interactions.items():
        if interactions:
            assert '{} {}'.format(len(interactions), itype) in found
        else:
            assert '{} {}'.format(0, itype) not in found

def test_to_molecule():
    """
    Test if the to molecule function gives
    expected results.
    """
    test_block = vermouth.molecule.Block()
    test_block.add_edges_from([('A', 'B'), ('B', 'C')])
    test_block.interactions["bonds"] = [
                Interaction(atoms=('A', 'B'),
                            parameters=['a', '0.2', '200'],
                            meta={'a': 0}),
                Interaction(atoms=('B', 'C'),
                            parameters=['a', '0.1', '300'],
                            meta={'b': 1}),]

    test_molecule = test_block.to_molecule()

    ref_bonds = [Interaction(atoms=(0, 1),
                             parameters=['a', '0.2', '200'],
                             meta={'a': 0}),
                 Interaction(atoms=(1, 2),
                             parameters=['a', '0.1', '300'],
                             meta={'b': 1}),]
    assert ref_bonds == test_molecule.interactions['bonds']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the :mod:`vermouth.edge_tuning` module.
"""

# The redefined-outer-name check from pylint wrongly catches the use of pytest
# fixtures.
# pylint: disable=redefined-outer-name

import copy
import pytest
import numpy as np
import networkx as nx
import vermouth
from vermouth import edge_tuning
from vermouth.molecule import Choice
from vermouth.utils import distance


@pytest.fixture
def molecule_for_pruning():
    """
    Build arbitrary graph to be pruned.
    """
    graph = nx.Graph([
        ['A', 'B'],
        ['C', 'A'],
        ['A', 'D'],
        ['B', 'D'],
        ['C', 'D'],
        ['E', 'B'],
        ['D', 'F'],
        ['E', 'F'],
        ['G', 'A'],
        ['G', 'C'],
    ])
    for key, value in graph.nodes.items():
        value['name'] = key
    return graph


@pytest.fixture
def simple_protein():
    """
    Build a protein-like molecule graph with possible cystein bridges.

    The molecule does not have coordinates.
    """
    graph = nx.Graph()
    graph.add_nodes_from((
        (0, {'atomname': 'BB', 'resname': 'CYS'}),
        (1, {'atomname': 'SG', 'resname': 'CYS'}),
        (2, {'atomname': 'BB', 'resname': 'OTHER'}),
        (3, {'atomname': 'SG', 'resname': 'OTHER'}),
        (4, {'atomname': 'BB', 'resname': 'CYS'}),
        (5, {'atomname': 'SG', 'resname': 'CYS'}),
        (6, {'atomname': 'BB', 'resname': 'CYS'}),
        (7, {'atomname': 'SG', 'resname': 'CYS'}),
        (8, {'atomname': 'BB', 'resname': 'CYS'}),
        (9, {'atomname': 'SG', 'resname': 'CYS'}),
        (10, {'atomname': 'BB', 'resname': 'CYS'}),
        (11, {'atomname': 'SG', 'resname': 'CYS'}),
        (12, {'atomname': 'BB', 'resname': 'CYS'}),
        (13, {'atomname': 'SG', 'resname': 'CYS'}),
        (14, {'atomname': 'BB', 'resname': 'OTHER'}),
        (15, {'atomname': 'SG', 'resname': 'OTHER'}),
        (16, {'atomname': 'BB', 'resname': 'OTHER'}),
        (17, {'atomname': 'SG', 'resname': 'OTHER'}),
    ))
    graph.add_edges_from((
        # Chain edges connecting the backbone and the side chains nodes
        (0, 1), (2, 3), (2, 4), (4, 5), (4, 6), (6, 7), (6, 8), (8, 9),
        (8, 10), (10, 11), (10, 12), (12, 13), (12, 14), (14, 15),
        (14, 16), (16, 17),
        # Bridges, including that should stay
        (1, 17), (3, 15), (5, 13), (7, 11),
    ))
    return graph


@pytest.fixture
def coordinate_array():
    """
    Build an array of coordinates for 36 points.

    The coordinates are random, but preset in the sense that they were rolled
    once and the same array is always returned.
    """
    # This array of coordinates was generated using:
    #    coordinates = (
    #        np.random.uniform(low=-2, high=2, size=(36, 3))
    #        .astype(np.float32)
    #        .round(2)
    #    )
    #    coordinates.tolist()
    return np.array([
        [0.5400000214576721, 1.690000057220459, 1.590000033378601],
        [1.7799999713897705, -0.10000000149011612, 0.41999998688697815],
        [1.9700000286102295, 1.409999966621399, 0.6000000238418579],
        [0.7300000190734863, 1.059999942779541, 1.1100000143051147],
        [1.3700000047683716, -2.0, 0.7799999713897705],
        [0.5899999737739563, 1.4299999475479126, 1.809999942779541],
        [-1.7400000095367432, -0.5400000214576721, 0.8500000238418579],
        [0.4300000071525574, -1.309999942779541, -1.7899999618530273],
        [-1.3300000429153442, 1.5399999618530273, -0.25],
        [0.8799999952316284, -1.9299999475479126, 1.2899999618530273],
        [1.7799999713897705, 1.5299999713897705, 1.690000057220459],
        [-1.7100000381469727, 0.20999999344348907, 0.8199999928474426],
        [0.07999999821186066, -1.1200000047683716, 1.399999976158142],
        [-0.3199999928474426, -0.49000000953674316, 0.9599999785423279],
        [0.7400000095367432, -1.8200000524520874, 1.1799999475479126],
        [1.0399999618530273, -1.7000000476837158, -1.6399999856948853],
        [0.4000000059604645, -1.0700000524520874, 1.25],
        [1.2200000286102295, 1.440000057220459, -0.6399999856948853],
        [-0.47999998927116394, -1.3799999952316284, 1.3200000524520874],
        [-0.46000000834465027, -0.07000000029802322, -0.5199999809265137],
        [0.05000000074505806, 0.5799999833106995, 1.659999966621399],
        [1.090000033378601, -0.17000000178813934, -0.8600000143051147],
        [1.159999966621399, 0.6499999761581421, 0.15000000596046448],
        [-0.3400000035762787, 0.9200000166893005, -0.07000000029802322],
        [0.17000000178813934, 1.2300000190734863, 0.23999999463558197],
        [1.559999942779541, 1.5800000429153442, -0.3100000023841858],
        [1.059999942779541, -0.6700000166893005, -0.4300000071525574],
        [-1.059999942779541, -0.3499999940395355, 1.9299999475479126],
        [-1.6299999952316284, 0.5299999713897705, -0.5299999713897705],
        [1.3799999952316284, -1.2699999809265137, 0.7400000095367432],
        [-1.4900000095367432, 1.6799999475479126, 0.41999998688697815],
        [1.159999966621399, -1.809999942779541, 1.2000000476837158],
        [-0.17000000178813934, -0.28999999165534973, 1.7899999618530273],
        [-1.9500000476837158, -1.25, 1.2200000286102295],
        [-0.05000000074505806, 0.8899999856948853, -1.350000023841858],
        [1.5099999904632568, 1.3300000429153442, 0.11999999731779099],
    ])


@pytest.fixture
def multi_molecules(coordinate_array):
    """
    Build 6 molecules with 6 atoms each. Coordinates are populated unter the
    `coords` attribute key with the coordinates produced by
    :func:`coordinate_array`.
    """
    molecules = []
    for _ in range(6):
        molecule = vermouth.molecule.Molecule()
        molecule.add_nodes_from([(idx, {'atomid': idx, 'resid': 1})
                                 for idx in range(6)])
        molecules.append(molecule)
    iter_nodes = (
        node
        for molecule in molecules
        for node in molecule.nodes.values()
    )
    for idx, (node, coords) in enumerate(zip(iter_nodes, coordinate_array)):
        node['coords'] = coords
        node['serial'] = idx
    return molecules


class TestPruneEdgesBetweenSelection:
    """
    Tests for :func:`edge_tuning.prune_edges_between_selections`.
    """
    @staticmethod
    @pytest.fixture
    def molecule_pruned(molecule_for_pruning):
        """
        Graph with edges pruned by
        :func:`edge_tuning.prune_edges_between_selections`.
        """
        graph = copy.deepcopy(molecule_for_pruning)
        selection_a = ['A', 'B', 'G']
        selection_b = ['C', 'D', 'G']
        edge_tuning.prune_edges_between_selections(
            graph, selection_a, selection_b
        )
        return graph

    @staticmethod
    @pytest.mark.parametrize('edge', [
        ['A', 'C'], ['A', 'D'], ['B', 'D']
    ])
    def test_prune_edges_between_selections_removed(molecule_pruned, edge):
        """
        Make sure that the edges pruned by
        :func:`edge_tuning.prune_edges_between_selections` are not present
        in the final graph.
        """
        assert edge not in molecule_pruned.edges

    @staticmethod
    @pytest.mark.parametrize('edge', [
        ['A', 'B'],  # Both in selection_a
        ['C', 'D'],  # Both in selection_b
        ['B', 'E'],  # E not in selections
        ['D', 'F'],  # F not in selections
        ['E', 'F'],  # None of E and F in selections
    ])
    def test_prune_edges_between_selections_kept(molecule_pruned, edge):
        """
        Make sure edges that should not be pruned by
        edge_tuning.prune_edges_between_selections` are still present in
        the final graph.
        """
        assert edge in molecule_pruned.edges


class TestPrudeEdgesWithSelectors:
    """
    Tests for :func:`edge_tuning.prune_edges_with_selectors`.
    """

    @staticmethod
    @pytest.fixture
    def molecule_pruned_one_selector(molecule_for_pruning):
        """
        Graph with edges pruned by :func:`edge_tuning.prune_edges_with_selectors`
        called with only :func:`dummy_selector_a`.
        """
        edge_tuning.prune_edges_with_selectors(molecule_for_pruning, dummy_selector_a)
        return molecule_for_pruning

    @staticmethod
    @pytest.fixture
    def molecule_pruned_two_selectors(molecule_for_pruning):
        """
        Graph with edges pruned by :func:`edge_tuning.prune_edges_with_selectors`
        called with both :func:`dummy_selector_a` and :func:`dummy_selector_b`.
        """
        edge_tuning.prune_edges_with_selectors(
            molecule_for_pruning, dummy_selector_a, dummy_selector_b
        )
        return molecule_for_pruning

    @staticmethod
    @pytest.mark.parametrize('edge', [
        ['A', 'B'], ['A', 'G'],
    ])
    def test_prune_edges_with_one_selector_removed(molecule_pruned_one_selector, edge):
        """
        Make sure that the edges pruned by
        :func:`edge_tuning.prune_edges_with_selectors` with a single selector
        provided are not present in the final graph.
        """
        assert edge not in molecule_pruned_one_selector.edges

    @staticmethod
    @pytest.mark.parametrize('edge', [
        ['A', 'C'], ['A', 'D'], ['B', 'D'],
    ])
    def test_prune_edges_with_two_selectors_removed(molecule_pruned_two_selectors, edge):
        """
        Make sure that the edges pruned by
        :func:`edge_tuning.prune_edges_with_selectors` with two selectors provided
        are not present in the final graph.
        """
        assert edge not in molecule_pruned_two_selectors.edges

    @staticmethod
    @pytest.mark.parametrize('edge', [
        ['A', 'B'],  # Both in selection_a
        ['C', 'D'],  # Both in selection_b
        ['B', 'E'],  # E not in selections
        ['D', 'F'],  # F not in selections
        ['E', 'F'],  # None of E and F in selections
    ])
    def test_prune_edges_with_two_selectors_kept(molecule_pruned_two_selectors, edge):
        """
        Make sure edges that should not be pruned by
        edge_tuning.prune_edges_with_selectors` with two selectors are still
        present in the final graph.
        """
        assert edge in molecule_pruned_two_selectors.edges


class TestAddEdgesAtDistance:
    """
    Tests for :func:`edge_tuning.add_edges_at_distance`.
    """
    @staticmethod
    @pytest.fixture
    def protein_with_coords(coordinate_array):
        """
        Build a graph with coordinates and edges placed by
        :func:`edge_tuning.add_edges_at_distance`.
        """
        graph = nx.Graph()
        graph.add_nodes_from([
            (index, {'coord': position})
            for index, position in enumerate(coordinate_array)
        ])
        edge_tuning.add_edges_at_distance(
            graph, 2.0, range(6), range(6, 37), attribute='coord'
        )
        return graph

    @staticmethod
    @pytest.mark.parametrize('edge', [
        (0, 10), (0, 20), (0, 22), (0, 24), (0, 35), (1, 16), (1, 17), (1, 21),
        (1, 22), (1, 25), (1, 26), (1, 29), (1, 31), (1, 35), (2, 10), (2, 17),
        (2, 22), (2, 24), (2, 25), (2, 35), (3, 10), (3, 13), (3, 17), (3, 20),
        (3, 22), (3, 23), (3, 24), (3, 25), (3, 32), (3, 35), (4, 9), (4, 12),
        (4, 14), (4, 16), (4, 26), (4, 29), (4, 31), (5, 10), (5, 20), (5, 22),
        (5, 24), (5, 32), (5, 35)
    ])
    def test_add_edges_at_distance(protein_with_coords, edge):
        """
        Make sure all the expected edges are added by
        :func:`edge_tuning.add_edges_at_distance`.
        """
        assert edge in protein_with_coords.edges

    @staticmethod
    def test_add_edges_at_distance_num(protein_with_coords):
        """
        Make sure that :func:`edge_tuning.add_edges_at_distance` adds the
        expected number of edges.

        This test passing is only relevant if :func:`test_add_edges_at_distance`
        passes as well.
        """
        assert len(protein_with_coords.edges) == 43


class TestAddInterMoleculeEdges:
    """
    Tests for :func:`add_inter_molecule_edges`.
    """
    @staticmethod
    @pytest.fixture
    def multi_molecules_linked(multi_molecules):
        """
        Merge the molecules from :func:`multi_molecules` using
        :func:`edge_tuning.add_inter_molecule_edges`.
        """
        edges = [
            ((0, 1), (4, 2)),
            ((4, 3), (5, 0)),
            ((2, 0), (3, 1)),
            ((3, 2), (2, 4)),
            ((1, 1), (1, 2)),
        ]
        return edge_tuning.add_inter_molecule_edges(multi_molecules, edges)

    @staticmethod
    def test_add_inter_molecule_edges_nmols(multi_molecules_linked):
        """
        Test that :func:`edge_tuning.add_inter_molecule_edges` produces
        the expected number of molecules.
        """
        print(multi_molecules_linked)
        assert len(multi_molecules_linked) == 3

    @staticmethod
    @pytest.mark.parametrize('mol, edge', (
        (0, (1, 8)),
        (0, (9, 12)),
        (2, (0, 7)),
        (2, (8, 4)),
        (1, (1, 2)),
    ))
    def test_add_inter_molecule_edges_edges(multi_molecules_linked, mol, edge):
        """
        Test that :func:`edge_tuning.add_inter_molecule_edges` creates the
        expected edges.
        """
        assert edge in multi_molecules_linked[mol].edges


class TestPairsUnderThreshold:
    """
    Tests for :func:`edge_tuning.pairs_under_threshold`.
    """
    @staticmethod
    @pytest.fixture
    def pair_selected(multi_molecules):
        """
        Call :func:`edge_tuning.pairs_under_threshold` with twice the same
        selection.
        """
        selection = [
            [0, 1],
            [0, 2],
            [1, 4],
            [2, 5],
            [3, 1],
            [5, 4],
        ]
        return edge_tuning.pairs_under_threshold(
            multi_molecules, 2.0, selection, selection, attribute='coords'
        )

    @staticmethod
    @pytest.fixture
    def assymetric_pair_selected(multi_molecules):
        """
        Call :func:`edge_tuning.pairs_under_threshold` with two different
        selections.
        """
        selection_a = [
            [0, 1],
            [1, 4],
            [3, 1],
        ]
        selection_b = [
            [0, 2],
            [2, 5],
            [5, 4],
        ]
        return edge_tuning.pairs_under_threshold(
            multi_molecules, 2.0, selection_a, selection_b, attribute='coords'
        )

    @staticmethod
    @pytest.mark.parametrize('edge', [
        ([0, 1], [0, 2]), ([0, 2], [1, 4]),
        ([0, 2], [2, 5]), ([3, 1], [5, 4]),
        ([0, 1], [2, 5]), ([2, 5], [5, 4]),
    ])
    def test_pairs_under_threshold_symetric(pair_selected, edge):
        """
        Test that :func:`edge_tuning.pairs_under_threshold` select the
        expected pairs when provided twice the same selection.
        """
        assert edge in list(pair[:2] for pair in pair_selected)

    @staticmethod
    def test_pairs_under_threshold_symetric_nedges(pair_selected):
        """
        Test that :func:`edge_tuning.pairs_under_threshold` select the
        expected number of pairs when provided twice the same selection.
        """
        # Each pair is yielded twice. Indeed, both selections are the same leading
        # to symetric pairs.
        assert len(list(pair_selected)) == 12

    @staticmethod
    def test_distance_under_threshold(multi_molecules, assymetric_pair_selected):
        distances = []
        found_distances = []
        for ((mol_a, key_a), (mol_b, key_b), dist) in assymetric_pair_selected:
            node_a = multi_molecules[mol_a].nodes[key_a]
            node_b = multi_molecules[mol_b].nodes[key_b]
            distances.append(distance(node_a['coords'], node_b['coords']))
            found_distances.append(dist)
        assert all(d <= 2.0 for d in distances)
        assert np.allclose(found_distances, distances)

    @staticmethod
    @pytest.mark.parametrize('edge', [
        ([0, 1], [0, 2]), ([1, 4], [0, 2]),
        ([3, 1], [5, 4]), ([0, 1], [2, 5]),
    ])
    def test_pairs_under_threshold_assymetric(assymetric_pair_selected, edge):
        """
        Test that :func:`edge_tuning.pairs_under_threshold` select the
        expected pairs when provided two different selections.
        """
        assert edge in list(pair[:2] for pair in assymetric_pair_selected)

    @staticmethod
    def test_pairs_under_threshold_assymetric_nedges(assymetric_pair_selected):
        """
        Test that :func:`edge_tuning.pairs_under_threshold` select the
        expected number of pairs when provided twice the same selection.
        """
        assert len(list(assymetric_pair_selected)) == 4

    @staticmethod
    def test_empty_selection(multi_molecules):
        """
        Make sure :func:`edge_tuning.pairs_under_threshold` is not failing on
        empty selections.
        """
        assert not list(edge_tuning.pairs_under_threshold(
            multi_molecules, 2.0, [], [], attribute='coords'
        ))


class TestAddEdgesThreshold:
    """
    Tests for :func:`edge_tuning.add_edges_threshold`.
    """
    @staticmethod
    @pytest.fixture
    def multi_molecules_with_edges(multi_molecules):
        """
        Creates multiple molecules connected using
        :func:`edge_tuning.add_edges_threshold`.
        """
        templates_a = [{'serial': Choice([1, 10, 19])}, {'name': 'not there'}]
        templates_b = [{'serial': Choice([2, 17])}, {'serial': 34}]
        return edge_tuning.add_edges_threshold(
            multi_molecules, 2.0, templates_a, templates_b, attribute='coords'
        )

    @staticmethod
    def test_add_edges_threshold_nmol(multi_molecules_with_edges):
        """
        Test that :func:`edge_tuning.add_edges_threshold` generate the expected
        number of molecules.
        """
        assert len(multi_molecules_with_edges) == 3

    @staticmethod
    def test_add_edges_threshold_nedges(multi_molecules_with_edges):
        """
        Test that :func:`edge_tuning.add_edges_threshold` generate the expected
        number of edges.
        """
        total_nedges = sum(len(molecule.edges) for molecule in multi_molecules_with_edges)
        assert total_nedges == 4


def test_select_nodes_multi(multi_molecules):
    """
    Test the output of :func:`edge_tuning.select_nodes_multi`.
    """
    selected = list(edge_tuning.select_nodes_multi(multi_molecules, dummy_selector))
    expected = [(molid, atomid) for molid in range(6) for atomid in range(0, 6, 2)]
    assert selected == expected


def dummy_selector_a(node):
    """
    Dummy node selector.
    """
    return node['name'] in ('A', 'B', 'G')


def dummy_selector_b(node):
    """
    Dummy node selector.
    """
    return node['name'] in ('C', 'D', 'G')


def dummy_selector(node):
    """
    Dummy selector that selects nodes with an even atomid.
    """
    return node['atomid'] % 2 == 0



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for the:class:`vermouth.processors.sort_molecule_atoms.SortMoleculeAtoms`
processor.
"""

import copy
from vermouth import Molecule, SortMoleculeAtoms
from vermouth.molecule import Interaction


def test_sort_molecule_atoms():
    """
    Test the :class:`vermouth.processors.sort_molecule_atoms.SortMoleculeAtoms`
    processor in normal conditions.
    """
    molecule = Molecule()
    nodes = [
        (6, {'chain': 'C', 'resid': 2, 'resname': 'AA0', 'atomname': 'A5'}),
        (5, {'chain': 'A', 'resid': 2, 'resname': 'XX0', 'atomname': 'A0'}),
        (4, {'chain': 'B', 'resid': 1, 'resname': 'XX1', 'atomname': 'A2'}),
        (3, {'chain': 'C', 'resid': 1, 'resname': 'XX0', 'atomname': 'A4'}),
        (2, {'chain': 'C', 'resid': 2, 'resname': 'AA1', 'atomname': 'A6'}),
        (1, {'chain': 'A', 'resid': 2, 'resname': 'XX1', 'atomname': 'A1'}),
        (0, {'chain': 'B', 'resid': 2, 'resname': 'XX0', 'atomname': 'A3'}),
    ]
    edges = [
        (0, 6, {'content': 'something'}),
        (1, 2, {'hello': 34}),
    ]
    interactions = {
        'bonds': [
            Interaction(atoms=[0, 1], parameters=['a', 'b'], meta={'a': 7}),
            Interaction(atoms=[3, 4], parameters=['c', 'd'], meta={'b': 5}),
        ],
        'angles': [
            Interaction(atoms=[0, 1, 2], parameters=['b', 'c'], meta={'a': 0}),
        ],
    }

    molecule.add_nodes_from(nodes)
    molecule.add_edges_from(edges)
    molecule.interactions = copy.copy(interactions)

    processor = SortMoleculeAtoms()
    processor.run_molecule(molecule)

    assert list(molecule.nodes) == [5, 1, 4, 0, 3, 6, 2]
    assert sorted(molecule.edges(data=True)) == edges
    assert molecule.interactions == interactions



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Read GROMACS .itp files.

"""

import collections
from vermouth.molecule import (Block, Interaction)
from vermouth.parser_utils import (SectionLineParser, _tokenize)

class ITPDirector(SectionLineParser):
    """
    class for reading itp files.
    """
    COMMENT_CHAR = ';'

    atom_idxs = {'bonds': [0, 1],
                 'position_restraints': [0],
                 'angles': [0, 1, 2],
                 'constraints': [0, 1],
                 'dihedrals': [0, 1, 2, 3],
                 'pairs': [0, 1],
                 'exclusions': [slice(None, None)],
                 'virtual_sitesn': [0, slice(2, None)],
                 'virtual_sites2': [0, 1, 2],
                 'virtual_sites3': [0, 1, 2, 3],
                 'pairs_nb': [0, 1],
                 'SETTLE': [0],
                 'virtual_sites4': [slice(0, 5)],
                 'distance_restraints':  [0, 1],
                 'dihedral_restraints':  [slice(0, 4)],
                 'orientation_restraints': [0, 1],
                 'angle_restraints': [slice(0, 4)],
                 'angle_restraints_z': [0, 1]}

    def __init__(self, force_field):
        super().__init__()
        self.force_field = force_field
        self.current_block = None
        self.current_meta = None
        self.blocks = collections.OrderedDict()
        self.header_actions = {
            ('moleculetype', ): self._new_block
        }

    def dispatch(self, line):
        """
        Looks at `line` to see what kind of line it is, and returns either
        :meth:`parse_header` if `line` is a section header or
        :meth:`vermouth.parser_utils.SectionLineParser.parse_section` otherwise.
        Calls :meth:`vermouth.parser_utils.SectionLineParser.is_section_header` to see
        whether `line` is a section header or not.

        Parameters
        ----------
        line: str

        Returns
        -------
        collections.abc.Callable
            The method that should be used to parse `line`.
        """

        if self.is_section_header(line):
            return self.parse_header
        elif self.is_pragma(line):
            return self.parse_pragma
        else:
            return self.parse_section

    @staticmethod
    def is_pragma(line):
        """
        Parameters
        ----------
        line: str
            A line of text.

        Returns
        -------
        bool
            ``True`` iff `line` is a def statement.
        """
        return line.startswith('#')

    def parse_pragma(self, line, lineno=0):
        """
        Parses the beginning and end of define sections
        with line number `lineno`. Sets attr current_meta
        when applicable. Does check if ifdefs overlap.

        Parameters
        ----------
        line: str
        lineno: str

        Returns
        -------
        object
            The result of calling :meth:`finalize_section`, which is called
            if a section ends.

        Raises
        ------
        IOError
            If the def sections are missformatted
        """

        if line == '#endif':
            if self.current_meta is not None:
                self.current_meta = None
            elif self.current_meta is None:
                raise IOError("Your #ifdef section is orderd incorrectly."
                              "At line {} I read #endif but I haven not read"
                              "a ifdef before.".format(lineno))
        elif line.startswith("#ifdef") or line.startswith("#ifndef"):
            if self.current_meta is None:
                condition, tag = line.split()
                self.current_meta = {'tag': tag, 'condition': condition.replace("#", "")}
            elif self.current_meta is not None:
                raise IOError("Your #ifdef/#ifndef section is orderd incorrectly."
                              "At line {} I read {} but there is still"
                              "an open #ifdef/#ifndef section from"
                              "before.".format(lineno, line.split()[0]))
        # Guard against unkown pragmas like #if or #include
        else:
            raise IOError("Don't know how to parse pargma {} at"
                          "line {}.".format(line, lineno))

    def parse_header(self, line, lineno=0):
        """
        Parses a section header with line number `lineno`. Sets
        :attr:`vermouth.parser_utils.SectionLineParser.section`
        when applicable. Does not check whether `line` is a valid section
        header.

        Parameters
        ----------
        line: str
        lineno: str

        Returns
        -------
        object
            The result of calling :meth:`finalize_section`, which is called
            if a section ends.

        Raises
        ------
        KeyError
            If the section header is unknown.
        """
        prev_section = self.section

        ended = []
        section = self.section + [line.strip('[ ]').casefold()]
        if tuple(section[-1:]) in self.METH_DICT:
            self.section = section[-1:]
        else:
            while tuple(section) not in self.METH_DICT and len(section) > 1:
                ended.append(section.pop(-2))  # [a, b, c, d] -> [a, b, d]
            self.section = section

        result = None

        if len(prev_section) != 0:
            result = self.finalize_section(prev_section, ended)

        action = self.header_actions.get(tuple(self.section))
        if action:
            action()

        return result

    def finalize_section(self, previous_section, ended_section):
        """
        Called once a section is finished. It appends the current_links list
        to the links and update the block dictionary with current_block. Thereby it
        finishes reading a given section.

        Parameters
        ---------
        previous_section: list[str]
            The last parsed section.
        ended_section: list[str]
            The sections that have been ended.
        """

        if self.current_block is not None:
            self.force_field.blocks[self.current_block.name] = self.current_block

    def finalize(self, lineno=0):
        """
        Called at the end of the file and checks that all pragmas are closed
        before calling the parent method.
        """
        if self.current_meta is not None:
            raise IOError("Your #ifdef/#ifndef section is orderd incorrectly."
                          "There is no #endif for the last pragma.")

        super().finalize()

    def _new_block(self):
        self.current_block = Block(force_field=self.force_field)

    @SectionLineParser.section_parser('moleculetype')
    def _block(self, line, lineno=0):
        """
        Parses the line directly following the '[moleculetype]'
        directive and stores the block name and exclusions.
         """
        name, nrexcl = line.split()
        self.current_block.name = name
        self.current_block.nrexcl = int(nrexcl)

    @SectionLineParser.section_parser('moleculetype', 'atoms')
    def _block_atoms(self, line, lineno=0):
        """
        Parses the lines of the [atoms] directive.
        """
        tokens = collections.deque(_tokenize(line))
        self._parse_block_atom(tokens, self.current_block)

    @SectionLineParser.section_parser('moleculetype', 'bonds')
    @SectionLineParser.section_parser('moleculetype', 'angles')
    @SectionLineParser.section_parser('moleculetype', 'dihedrals')
    @SectionLineParser.section_parser('moleculetype', 'impropers')
    @SectionLineParser.section_parser('moleculetype', 'constraints')
    @SectionLineParser.section_parser('moleculetype', 'pairs')
    @SectionLineParser.section_parser('moleculetype', 'exclusions')
    @SectionLineParser.section_parser('moleculetype', 'virtual_sites2')
    @SectionLineParser.section_parser('moleculetype', 'virtual_sites3')
    @SectionLineParser.section_parser('moleculetype', 'virtual_sites4')
    @SectionLineParser.section_parser('moleculetype', 'virtual_sitesn')
    @SectionLineParser.section_parser('moleculetype', 'position_restraints')
    @SectionLineParser.section_parser('moleculetype', 'pairs_nb')
    @SectionLineParser.section_parser('moleculetype', 'SETTLE')
    @SectionLineParser.section_parser('moleculetype', 'distance_restraints')
    @SectionLineParser.section_parser('moleculetype', 'orientation_restraints')
    @SectionLineParser.section_parser('moleculetype', 'angle_restraints')
    @SectionLineParser.section_parser('moleculetype', 'angle_restraints_z')
    def _interactions(self, line, lineno=0):
        """
        Parses all interaction lines that are not directives (i.e. within []).
        Note that each interaction is enumerated explicitly to guard against typos
        and also interactions for which the format is unknown.
        """
        context = self.current_block
        interaction_name = self.section[-1]
        tokens = collections.deque(_tokenize(line))

        atom_idxs = self.atom_idxs.get(interaction_name)

        self._base_parser(
            tokens,
            context,
            section=interaction_name,
            atom_idxs=atom_idxs,
            )

    def _split_atoms_and_parameters(self, tokens, atom_idxs):
        """
        Returns atoms from line based on the indices defined in `atom_idxs`.
        It also interprets slices etc. stored as strings.

        Parameters:
        ------------
        tokens: collections.deque[str]
            Deque of token to inspect. The deque **can be modified** in place.
        atom_idxs: list of ints or strings that are valid python slices

        Returns:
        -----------
        list
        """

        atoms = []
        remove = []
        # first we extract the atoms from the indices given using
        # ints or slices
        tokens = list(tokens)
        for idx in atom_idxs:
            if isinstance(idx, int):
                atoms.append([tokens[idx], {}])
                remove.append(idx)
            elif isinstance(idx, slice):
                atoms += [[atom, {}] for atom in tokens[idx]]
                idx_range = range(0, len(tokens))
                remove += idx_range[idx]
            else:
                raise IOError

        # everything that is left are parameters, which we
        # get by simply deleting the atoms from tokens

        for index in sorted(remove, reverse=True):
            del tokens[index]

        return atoms, tokens

    def _treat_block_interaction_atoms(self, atoms, context, section):
        """
        Takes the atom indices associated with an interaction line
        and converts it to zero based indices. It also performas some
        format checks. It checks that:

              - the indices are not negative or zero
              - the indices refere to an existing atom
              - atoms have no prefixes

        Parameters
        -----------
        atom_idxs: list of ints or strings that are valid python slices
        context: :class:`vermouth.molecule.Block`
            The current block we parse
        section: str
            The current section header
        """
        atom_names = list(context.nodes)
        all_references = []
        for atom in atoms:
            reference = atom[0]
            if reference.isdigit():
                if int(reference) < 1:
                    msg = 'In section {} is a negative atom reference, which is not allowed.'
                    raise IOError(msg.format(section.name))

               # The indices in the file are 1-based
                reference = int(reference) - 1
                try:
                    reference = atom_names[reference]
                except IndexError:
                    msg = ('There are {} atoms defined in the block "{}". '
                           'Interaction in section "{}" cannot refer to '
                           'atom index {}.')
                    raise IOError(msg.format(len(context), context.name,
                                             section, reference + 1))
                atom[0] = reference
            else:
                if reference not in context:
                    msg = ('There is no atom "{}" defined in the block "{}". '
                           'Section "{}" cannot refer to it.')
                    raise IOError(msg.format(reference, context.name, section))
                if reference[0] in '+-<>':
                    msg = ('Atom names in blocks cannot be prefixed with + or -. '
                           'The name "{}", used in section "{}" of the block "{}" '
                           'is not valid in a block.')
                    raise IOError(msg.format(reference, section, context.name))
            all_references.append(reference)
        return all_references

    def _base_parser(self, tokens, context, section, atom_idxs):
        """
        Converts an interaction line into a vermouth interaction
        tuple. It updates the block interactions in place.

        Parameters
        ----------
        tokens: collections.deque[str]
            Deque of token to inspect. The deque **can be modified** in place.
        context: :class:`vermouth.molecule.Block`
            The current block we parse
        section: str
            The current section header
        atom_idxs: list of ints or strings that are valid python slices
        """
        # split atoms and parameters

        atoms, parameters = self._split_atoms_and_parameters(tokens, atom_idxs)

        # perform check on the atom ids
        treated_atoms = self._treat_block_interaction_atoms(atoms, context, section)

        if self.current_meta:
            meta = {self.current_meta['condition']: self.current_meta['tag']}
        else:
            meta = {} #dict(collections.ChainMap(meta, apply_to_all_interactions))

        interaction = Interaction(
            atoms=treated_atoms,
            parameters=parameters,
            meta=meta,)

        context.interactions[section] = context.interactions.get(section, []) + [interaction]

    def _parse_block_atom(self, tokens, context):
        """
        Converts the lines of the atom directive to graph nodes and
        sets the values (i.e. atomtype) as attributes.

        Parameters
        ----------
        tokens: collections.deque[str]
            Deque of token to inspect. The deque **can be modified** in place.
        context: :class:`vermouth.molecule.Block`
            The current block we parse
        """
        # deque does not support slicing
        first_six = (tokens.popleft() for _ in range(6))
        index, atype, resid, resname, name, charge_group = first_six
        # since the index becomes the node name and all graphs start with 0
        # index it makes more sense to also directly start at 0

        if int(index) < 1:
            msg = 'One of your atoms has a negative atom reference, which is not allowed.'
            raise IOError(msg)

        index = int(index) - 1

        if index in context:
            msg = ('There is already an atom with index "{}" in the block "{}". '
                   'Atom indices must be unique within a block.')
            raise IOError(msg.format(name, context.name))

        atom = {
            # for bookkeeping purposes let's keep the actual index i.e. +1 here
            'index'   : index + 1,
            'atomname': name,
            'atype': atype,
            'resname': resname,
            'resid': int(resid),
            'charge_group': int(charge_group),
        }

       # charge and mass are optional, but charge has to be defined for mass to be
        if tokens:
            atom['charge'] = float(tokens.popleft())
        if tokens:
            atom['mass'] = float(tokens.popleft())

        attributes = {}

        context.add_node(index, **dict(collections.ChainMap(attributes, atom)))

def read_itp(lines, force_field):
    """
    Parses `lines` of itp format and adds the
    molecule as a block to `force_field`.

    Parameters
    ----------
    lines: list
        list of lines of an itp file
    force_field: :class:`vermouth.forcefield.ForceField`
    """
    director = ITPDirector(force_field)
    return list(director.parse(iter(lines)))
